'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from collections import Counter
from typing import Optional, Dict, Any

from pm4py.algo.discovery.inductive.fall_through.tau_loop import TauLoopUVCL
from pm4py.algo.discovery.powl.inductive.fall_through.strict_tau_loop import (
    POWLStrictTauLoopUVCL,
)
from pm4py.util.compression import util as comut
from pm4py.util.compression.dtypes import UVCL


class POWLTauLoopUVCL(POWLStrictTauLoopUVCL, TauLoopUVCL):

    @classmethod
    def _get_projected_log(
        cls, log: UVCL, parameters: Optional[Dict[str, Any]] = None
    ) -> UVCL:
        start_activities = comut.get_start_activities(log)
        proj = Counter()
        for t in log:
            x = 0
            for i in range(1, len(t)):
                if t[i] in start_activities:
                    proj.update({t[x:i]: log[t]})
                    x = i
            proj.update({t[x: len(t)]: log[t]})
        return proj



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.util import constants, xes_constants, exec_utils
from enum import Enum
from collections import Counter
from pm4py.objects.log.obj import EventLog, EventStream
from typing import Union, Optional, Dict, Any
from pm4py.objects.conversion.log import converter as log_converter


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def apply(
    log: Union[EventLog, EventStream],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Dict[str, int]:
    """
    Associates to each activity (with at least one rework) the number of cases in the log for which
    the rework happened.

    Parameters
    ------------------
    log
        Event log
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => the attribute to be used as activity

    Returns
    ------------------
    dict
        Dictionary associating to each activity the number of cases for which the rework happened
    """
    if parameters is None:
        parameters = {}

    log = log_converter.apply(log, variant=log_converter.Variants.TO_EVENT_LOG)

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    ret = Counter()

    for trace in log:
        activities = Counter([x[activity_key] for x in trace])
        activities = [x for x in activities if activities[x] > 1]
        for act in activities:
            ret[act] += 1

    return dict(ret)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any, Set, Tuple


def apply(
    ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None
) -> Set[Tuple[str, str]]:
    """
    Calculates the object interaction graph. Two objects are connected iff they are both related to an event
    of the OCEL.

    Parameters
    -----------------
    ocel
        Object-centric event log
    parameters
        Parameters of the algorithm

    Returns
    -----------------
    object_interaction_graph
        Object interaction graph (as set of tuples; undirected)
    """
    if parameters is None:
        parameters = {}

    graph = set()

    ev_rel_obj = (
        ocel.relations.groupby(ocel.event_id_column)[ocel.object_id_column]
        .agg(list)
        .to_dict()
    )

    for ev in ev_rel_obj:
        rel_obj = ev_rel_obj[ev]
        for o1 in rel_obj:
            for o2 in rel_obj:
                if o1 < o2:
                    graph.add((o1, o2))

    return graph



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from enum import Enum
import pandas as pd
from pm4py.objects.ocel.obj import OCEL
from pm4py.objects.ocel import constants
from pm4py.util import exec_utils


class Parameters(Enum):
    EVENT_ID = constants.PARAM_EVENT_ID
    EVENT_ACTIVITY = constants.PARAM_EVENT_ACTIVITY
    EVENT_TIMESTAMP = constants.PARAM_EVENT_TIMESTAMP
    OBJECT_ID = constants.PARAM_OBJECT_ID
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE
    OBJECTS_UNIQUE_PER_TRACE = "objects_unique_per_trace"

def feasible_traces_to_ocel(
    feasible_traces_iter, all_transitions, id_to_obj_type, parameters
):
    """
    Converts the feasible traces into an OCEL object.

    Parameters
    ----------
    feasible_traces_iter : iter
        An iterator over feasible traces, where each trace is a tuple of events
    all_transitions : list
        Ordered list of all transitions in the object-centric Petri net
    id_to_obj_type : dict
        Mapping from object IDs to their types
    parameters : dict
        Additional parameters for the conversion

    Returns
    -------
    OCEL
        The resulting OCEL object
    """
    event_id_column = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, constants.DEFAULT_EVENT_ID
    )
    object_id_column = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, constants.DEFAULT_OBJECT_ID
    )
    object_type_column = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, constants.DEFAULT_OBJECT_TYPE
    )

    event_activity = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, constants.DEFAULT_EVENT_ACTIVITY
    )
    event_timestamp = exec_utils.get_param_value(
        Parameters.EVENT_TIMESTAMP, parameters, constants.DEFAULT_EVENT_TIMESTAMP
    )
    objects_unique_per_trace = exec_utils.get_param_value(
        Parameters.OBJECTS_UNIQUE_PER_TRACE, parameters, False
    )
    # Convert all found traces to OCEL format

    # Create the OCEL object
    events_list = []
    objects_list = []
    relations_list = []

    all_objects_seen = set()
    event_id_counter = 0
    # assigns to each event an increased timestamp from 1970
    curr_timestamp = 10000000
    
    if objects_unique_per_trace:
        object_id_counter = 0

    # convert trace to set of events, objects, and relations
    for trace in feasible_traces_iter:
        for event in trace:
            transition_idx, object_ids = event
            transition = all_transitions[transition_idx]

            # Do not add silent transitions to the log
            if transition.label is not None:
                # Create event
                event_id = f"event_{event_id_counter}"
                event_id_counter += 1
                curr_timestamp += 1

                events_list.append(
                    {
                        event_id_column: event_id,
                        event_activity: transition.label,
                        event_timestamp: pd.to_datetime(curr_timestamp, unit="s"),
                    }
                )

                # Create objects and relations
                for obj_id in object_ids:
                    obj_type = id_to_obj_type[obj_id]
                    if objects_unique_per_trace:
                            obj_id = f"{obj_id}_{object_id_counter}"

                    # Add object
                    if obj_id not in all_objects_seen:
                        all_objects_seen.add(obj_id)
                        objects_list.append(
                            {object_id_column: obj_id, object_type_column: obj_type}
                        )

                    # Add relation
                    relations_list.append(
                        {
                            event_id_column: event_id,
                            event_activity: transition.label,
                            event_timestamp: pd.to_datetime(curr_timestamp, unit="s"),
                            object_id_column: obj_id,
                            object_type_column: obj_type,
                        }
                    )
        if objects_unique_per_trace:
            object_id_counter += 1

    # Convert to dfs
    events_df = pd.DataFrame(events_list)
    objects_df = pd.DataFrame(objects_list)
    relations_df = pd.DataFrame(relations_list)

    # Create the OCEL object
    ocel = OCEL(
        events=events_df,
        objects=objects_df,
        relations=relations_df,
        parameters=parameters,
    )

    return ocel


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import tempfile
from copy import deepcopy, copy
from enum import Enum

from graphviz import Graph

from pm4py.objects.process_tree.utils import generic
from pm4py.util import exec_utils
from typing import Optional, Dict, Any, Union
from pm4py.objects.process_tree.obj import ProcessTree
import graphviz
from pm4py.util import vis_utils, constants


class Parameters(Enum):
    FORMAT = "format"
    ENABLE_DEEPCOPY = "enable_deepcopy"
    FONT_SIZE = "font_size"
    BGCOLOR = "bgcolor"
    RANKDIR = "rankdir"
    NUM_EVENTS_PROPERTY = "num_events_property"
    NUM_CASES_PROPERTY = "num_cases_property"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


# maps the operators to the ProM strings
operators_mapping = {
    "->": "seq",
    "X": "xor",
    "+": "and",
    "*": "xor loop",
    "O": "or",
    "<>": "interleaving",
}

# root node parameter
ROOT_NODE_PARAMETER = "@@root_node"


def repr_tree_2(tree, viz, parameters):
    num_events_property = exec_utils.get_param_value(
        Parameters.NUM_EVENTS_PROPERTY, parameters, "num_events"
    )
    num_cases_property = exec_utils.get_param_value(
        Parameters.NUM_CASES_PROPERTY, parameters, "num_cases"
    )
    root_node = parameters[ROOT_NODE_PARAMETER]

    root_node_num_cases = root_node._properties[num_cases_property]
    this_node_num_cases = (
        tree._properties[num_cases_property]
        if num_cases_property in tree._properties
        else 0
    )
    this_node_num_events = (
        tree._properties[num_events_property]
        if num_events_property in tree._properties
        else 0
    )

    font_size = exec_utils.get_param_value(
        Parameters.FONT_SIZE, parameters, 15
    )
    font_size = str(font_size)

    this_node_id = str(id(tree))

    if tree.operator is None:
        if tree.label is None:
            viz.node(
                this_node_id,
                "tau",
                style="filled",
                fillcolor="black",
                shape="point",
                width="0.075",
                fontsize=font_size,
            )
        else:
            node_color = vis_utils.get_trans_freq_color(
                this_node_num_cases, 0, root_node_num_cases
            )
            node_label = str(tree) + "\nC=%d E=%d" % (
                this_node_num_cases,
                this_node_num_events,
            )
            viz.node(
                this_node_id,
                node_label,
                fontsize=font_size,
                style="filled",
                fillcolor=node_color,
            )
    else:
        node_color = vis_utils.get_trans_freq_color(
            this_node_num_cases, 0, root_node_num_cases
        )
        viz.node(
            this_node_id,
            operators_mapping[str(tree.operator)],
            fontsize=font_size,
            style="filled",
            fillcolor=node_color,
        )

        for child in tree.children:
            repr_tree_2(child, viz, parameters)

    if tree.parent is not None:
        viz.edge(str(id(tree.parent)), this_node_id, dirType="none")


def apply(
    tree: ProcessTree,
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> graphviz.Graph:
    """
    Obtain a Process Tree representation through GraphViz

    Parameters
    -----------
    tree
        Process tree
    parameters
        Possible parameters of the algorithm

    Returns
    -----------
    gviz
        GraphViz object
    """
    if parameters is None:
        parameters = {}

    parameters = copy(parameters)
    parameters[ROOT_NODE_PARAMETER] = tree

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )
    rankdir = exec_utils.get_param_value(
        Parameters.RANKDIR, parameters, constants.DEFAULT_RANKDIR_GVIZ
    )
    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Process Tree"
    )
    font_size = exec_utils.get_param_value(
        Parameters.FONT_SIZE, parameters, 15
    )
    font_size = str(font_size)

    viz = Graph(
        "pt",
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor, "rankdir": rankdir},
    )
    viz.attr("node", shape="ellipse", fixedsize="false")

    if enable_graph_title:
        viz.attr(
            label='<<FONT POINT-SIZE="'
            + str(2 * int(font_size))
            + '">'
            + graph_title
            + "</FONT>>",
            labelloc="top",
        )

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )

    enable_deepcopy = exec_utils.get_param_value(
        Parameters.ENABLE_DEEPCOPY, parameters, False
    )

    if enable_deepcopy:
        # since the process tree object needs to be sorted in the visualization, make a deepcopy of it before
        # proceeding
        tree = deepcopy(tree)
        generic.tree_sort(tree)

    repr_tree_2(tree, viz, parameters)

    viz.attr(overlap="false")
    viz.attr(splines="false")
    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Set


def pre_set(elem, object_type: str = None) -> Set:
    """
    Returns the set of predecessors of an element (place or transition) in an object-centric Petri net.
    Restricted to predecessors connected using arcs of the object type, if specified.
    
    Parameters
    ----------
    elem
        Element (place or transition) for which to get the predecessors
    object_type
        Object type to restrict the predecessors to (optional)
    
    Returns
    -------
    Set
        Set of predecessor elements (places or transitions) of the specified object type.
    """
    pre = set()
    for a in elem.in_arcs:
        if object_type is None or a.object_type == object_type:
            pre.add(a.source)
    return pre

def post_set(elem, object_type: str = None) -> Set:
    """
    Returns the set of successors of an element (place or transition) in an object-centric Petri net.
    Restricted to successors connected using arcs of the object type, if specified.

    Parameters
    ----------
    elem
        Element (place or transition) for which to get the successors
    object_type
        Object type to restrict the successors to (optional)
    
    Returns
    -------
    Set
        Set of successor elements (places or transitions) of the specified object type.
    """
    post = set()
    for a in elem.out_arcs:
        if object_type is None or a.object_type == object_type:
            post.add(a.target)
    return post


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.algo.transformation.to_embeddings.variants import cases_transformers, events_transformers
from enum import Enum
from pm4py.util import exec_utils
import pandas as pd
from typing import Optional, Dict, Any, List, Tuple


class Variants(Enum):
    CASES_TRANSFORMERS = cases_transformers
    EVENTS_TRANSFORMERS = events_transformers


def apply(log: pd.DataFrame, variant=Variants.CASES_TRANSFORMERS, parameters: Optional[Dict[Any, Any]] = None) -> Tuple[
    List[str], List[List[float]]]:
    """
    Computes the embeddings (case/event level, depending on the variant) of the provided dataframe.

    Parameters
    -----------------
    log
        Pandas dataframe
    variant
        Variant of the algorithm, including:
        - Variants.CASES_TRANSFORMERS => computes the embeddings at the case level
        - Variants.EVENTS_TRANSFORMERS => computes the embeddings at the event level
    parameters
        Variant-specific parameters

    Returns
    ----------------
    ids
        Identifiers of the considered events/cases
    embeddings_list
        List of embeddings for the considered events/cases
    """
    return exec_utils.get_variant(variant).apply(log, parameters=parameters)


def keep_top_k_per_similarity(log: pd.DataFrame, target_sentence: str, k: int, variant=Variants.CASES_TRANSFORMERS,
                              parameters: Optional[Dict[Any, Any]] = None) -> pd.DataFrame:
    """
    Keeps the top K events/cases per similarity

    Parameters
    ----------------
    log
        Pandas dataframe
    variant
        Variant of the algorithm, including:
        - Variants.CASES_TRANSFORMERS => computes the embeddings at the case level
        - Variants.EVENTS_TRANSFORMERS => computes the embeddings at the event level
    parameters
        Variant-specific parameters

    Returns
    -----------------
    filtered_log
        Filtered event log
    """
    return exec_utils.get_variant(variant).keep_top_k_per_similarity(log, target_sentence, k, parameters=parameters)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or 
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from collections import Counter
import uuid
from pm4py.objects.ocpn.obj import OCMarking, OCPetriNet
from typing import Dict, Any

from pm4py.objects.petri_net.obj import PetriNet


def create(ocpn: Dict[str, Any]) -> OCPetriNet:
    """
    Creates an Object-centric Petri net object from its dictionary representation
    specified in pm4py.algo.discovery.ocel.ocpn.variants.classic.
    Only considers the properties `activities`, `petri_nets`, and `double_arcs_on_activity`.
    All other information is ignored.
    
    Parameters
    -----------------
    ocpn
        Dictionary containing the properties of the Object-centric Petri net

    Returns
    ----------------
    OCPetriNet
        Object-centric Petri net object
    """
    activities = ocpn["activities"]
    petri_nets = ocpn["petri_nets"]
    double_arcs_on_activity = ocpn["double_arcs_on_activity"]
    
    places = dict()
    unlabeled_transitions = dict()
    arcs = []
    initial_marking = OCMarking()
    final_marking = OCMarking()
    
    # Labeled transitions
    labeled_transitions = {label: OCPetriNet.Transition(label=label, name=str(uuid.uuid4())) for label in activities}
    
    for ot, net in petri_nets.items():
        pn, im, fm = net
        
        # transitions
        for t in pn.transitions:
            if not t.label:
                # labeled transitions are already in labeled_transitions
                name = f"{ot}_{t.name}" # make name unique
                unlabeled_transitions[name] = OCPetriNet.Transition(name=name)
        
        # places 
        for p in pn.places:
            name = f"{ot}_{p.name}" # make name unique
            places[name] = OCPetriNet.Place(name=name, object_type=ot)
            
        # arcs
        for arc in pn.arcs:
            is_double = False
            if isinstance(arc.source, PetriNet.Transition):
                if arc.source.label:
                    source = labeled_transitions[arc.source.label]
                    is_double = double_arcs_on_activity[ot][arc.source.label]
                else:
                    source = unlabeled_transitions[f"{ot}_{arc.source.name}"]
                target = places[f"{ot}_{arc.target.name}"]
            elif isinstance(arc.source, PetriNet.Place):
                source = places[f"{ot}_{arc.source.name}"]
                if arc.target.label:
                    target = labeled_transitions[arc.target.label]
                    is_double = double_arcs_on_activity[ot][arc.target.label]
                else:
                    target = unlabeled_transitions[f"{ot}_{arc.target.name}"]
            else:
                raise ValueError("Unknown arc source type")
            
            # check for double arc
        
            a = OCPetriNet.Arc(source=source, target=target, object_type=ot, is_variable=is_double)
            arcs.append(a)
            source.add_out_arc(a)
            target.add_in_arc(a)
        
        # markings
        for p in im:
            initial_marking += OCMarking({places[f"{ot}_{p.name}"]: Counter([f"{ot}_{i}" for i in range(im[p])])})
        for p in fm:
            final_marking += OCMarking({places[f"{ot}_{p.name}"]: Counter([f"{ot}_{i}" for i in range(fm[p])])})
    
    # create the OCPetriNet object
    ocpn_obj = OCPetriNet(
        places = set(places.values()),
        transitions = set(labeled_transitions.values()) | set(unlabeled_transitions.values()),
        arcs = set(arcs),
        initial_marking=initial_marking,
        final_marking=final_marking,
    )
    return ocpn_obj


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocpn.variants import to_oc_causal_net, to_alternative_format
from pm4py.util import exec_utils
from enum import Enum

class Variants(Enum):
    TO_OC_CAUSAL_NET = to_oc_causal_net
    TO_ALTERNATIVE_FORMAT =  to_alternative_format
    
def apply(ocpn, parameters=None, variant=Variants.TO_OC_CAUSAL_NET):
    """
    Method for converting an Object-centric Petri Net to Object-centric Causal Net

    Parameters
    -----------
    ocpn: OCPetriNet
        Object-centric Petri net
    parameters: dict, optional
        Parameters of the algorithm
    variant
        Chosen variant of the algorithm

    Returns
    -----------
    Conversion result
    """
    return exec_utils.get_variant(variant).apply(ocpn, parameters=parameters)


"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from collections import Counter, namedtuple
import random
from typing import Any, Dict, Optional

import pandas as pd
from pm4py.objects.oc_causal_net.obj import OCCausalNet
from pm4py.objects.oc_causal_net.semantics import OCCausalNetState, OCCausalNetSemantics
from pm4py.objects.ocel import constants
from pm4py.objects.ocel.obj import OCEL
from pm4py.util import exec_utils
from enum import Enum


class Parameters(Enum):
    EVENT_ID = constants.PARAM_EVENT_ID
    EVENT_ACTIVITY = constants.PARAM_EVENT_ACTIVITY
    EVENT_TIMESTAMP = constants.PARAM_EVENT_TIMESTAMP
    OBJECT_ID = constants.PARAM_OBJECT_ID
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE
    OBJECTS_UNIQUE_PER_SEQUENCE = "objects_unique_per_sequence"
    RETURN_SEQUENCES = "return_sequences"
    MAX_BINDINGS_PER_ACTIVITY = "maxBindingsPerActivity"
    OCCN_SEMANTICS = "occn_semantics"
    BRANCHING_FACTOR_ACTIVITIES = "branching_factor_activities"
    BRANCHING_FACTOR_BINDINGS = "branching_factor_bindings"


FINAL_MARKER = "FINAL"


# Define memory-efficient data type for binding
# a binding is a tuple of activity id, consumed objects, and produced objects
#    consumed / produces are tuples of (predecessor/successor activity id, objects_per_ot)
#      where objects_per_ot is a tuples of entries (object_type, objects)
#         where objects is a tuple (obj_id_1, obj_id_2, ...)
Binding = namedtuple("Binding", ["activity_id", "consumed", "produced"])


def apply(
    occn: OCCausalNet, objects: dict, parameters: Optional[Dict[Any, Any]] = None
) -> OCEL:
    """
    Compute playout of an object-centric causal net generating an OCEL.
    Extensive search, retrieves all valid binding sequences.
    Starts by binding start activities with the objects specified and ends in the empty state.
    The empty sequence is considered a valid sequence no other sequences are found.

    Parameters
    -----------
    occn
        Object-centric causal net to play-out
    objects
        Dictionary mapping object types to sets of object ids. These objects will be introduced by the start activities of the occn at the beginning of every binding sequence.
    parameters
        Parameters of the algorithm, including:
            Parameters.MAX_BINDINGS_PER_ACTIVITY: Maximum number of bindings per activity (mandatory)
            Parameters.RETURN_SEQUENCES: If True, return an iterator to all possible sequences of bindings instead of an OCEL
            Parameters.OBJECTS_UNIQUE_PER_SEQUENCE: If True, objects in the resulting OCEL are make unique per sequence (default: False)
            Parameters.OCCN_SEMANTICS: The semantics to be used for the causal net (default: OCCausalNetSemantics())
            Parameters.BRANCHING_FACTOR_ACTIVITIES: Maximum branching factor for exploring enabled activities (default: inf). Note that the play-out will generate a subset of all sequences if this is set.
            Parameters.BRANCHING_FACTOR_BINDINGS: Maximum branching factor for exploring enabled bindings (default: inf). Note that the play-out will generate a subset of all sequences if this is set.
    """
    if parameters is None:
        parameters = {}

    return_sequences = exec_utils.get_param_value(
        Parameters.RETURN_SEQUENCES, parameters, False
    )
    if Parameters.MAX_BINDINGS_PER_ACTIVITY not in parameters:
        raise ValueError(
            "Parameter MAX_BINDINGS_PER_ACTIVITY must be specified for the extensive playout. This parameter limits the maximum number of times an activity may be executed."
        )
    max_bindings_per_activity = exec_utils.get_param_value(
        Parameters.MAX_BINDINGS_PER_ACTIVITY, parameters, None
    )
    semantics = exec_utils.get_param_value(
        Parameters.OCCN_SEMANTICS,
        parameters,
        OCCausalNetSemantics(),
    )
    bf_act = exec_utils.get_param_value(
        Parameters.BRANCHING_FACTOR_ACTIVITIES, parameters, float("inf")
    )
    bf_bind = exec_utils.get_param_value(
        Parameters.BRANCHING_FACTOR_BINDINGS, parameters, float("inf")
    )

    # create int id for every activity for memory efficiency
    activity_to_id = {activity: i for i, activity in enumerate(occn.activities)}
    id_to_activity = {i: activity for activity, i in activity_to_id.items()}
    start_activities = set(
        i for activity, i in activity_to_id.items() if activity.startswith("START_")
    )
    # same for object types
    object_type_to_id = {
        object_type: i for i, object_type in enumerate(occn.object_types)
    }
    id_to_object_type = {i: object_type for object_type, i in object_type_to_id.items()}

    # Set up initial state with starting objects
    # In the state, we denote activities by their id, not by their name
    initial_state = OCCausalNetState()

    # Create fake obligations to start activities for all starting objects
    for object_type, object_ids in objects.items():
        ot_id = object_type_to_id[object_type]
        start_activity_id = activity_to_id[f"START_{object_type}"]
        initial_state += OCCausalNetState(
            {start_activity_id: Counter([(-1, obj_id, ot_id) for obj_id in object_ids])}
        )

    # Activity counts
    # index is from `activity_to_id`
    initial_activity_counts = (0,) * len(occn.activities)

    # State key used for memoization, see below
    initial_state_key = (initial_state, initial_activity_counts)

    # Memoization cache: Dict[state_key, Union[Set[Tuple[Binding, next_key]], str]]
    # where state_key is a tuple of (state, activity_counts) and the value is either
    # FINAL_MARKER if the state is the empty state,
    # or a set of tuples of bindings and next state keys that correspond to all successor
    # states that can be reached from the current state using the respective bindings.
    memo = {}

    # == Phase 1: Memoization DFS Graph Population ==
    _populate_memo_graph(
        initial_state_key,
        occn,
        semantics,
        max_bindings_per_activity,
        start_activities,
        activity_to_id,
        id_to_activity,
        object_type_to_id,
        bf_act,
        bf_bind,
        memo,
    )

    # == Phase 2: Reconstruct traces from memo ==
    valid_sequences_iter = _reconstruct_sequences(initial_state_key, memo)

    # == Phase 3: Return data in the desired format ==
    if return_sequences:
        # return valid_sequences along with a mapping from indices to activities and object types
        return (
            valid_sequences_iter,
            id_to_activity,
            id_to_object_type,
        )
    else:
        return _valid_sequences_to_ocel(valid_sequences_iter, id_to_activity, id_to_object_type, parameters)


def _populate_memo_graph(
    state_key: tuple,
    occn: OCCausalNet,
    semantics,
    max_bindings: int,
    start_activities,
    act_to_idx: dict,
    idx_to_act: dict,
    ot_to_idx: dict,
    bf_act: float,
    bf_bind: float,
    memo: dict,
) -> bool:
    """
    Recursively explores the state space to build a compact, memoized graph of all valid binding sequences.

    This function performs a depth-first search from a given state_key. It populates a memoization
    cache (`memo`). For each state (defined by the state_key), it stores the set of "next steps" (as tuples of
    (binding, next_state_key)) that lie on a path to the empty state,
    where binding is of type Binding.

    This approach avoids duplicate computation of two different sequences leading to the same state key.

    Parameters
    -----------
    state_key : tuple
        A tuple representing the current state in the form (state, activity_counts).
    occn : OCCausalNet
        The object-centric causal net being used.
    semantics
        The semantics to be used for the causal net.
    max_bindings : int
        Maximum number of bindings per activity.
    start_activities
        Collection of indices for start activities.
    act_to_idx : dict
        Dictionary mapping activities to their id.
    idx_to_act : dict
        Dictionary mapping activity ids to their names.
    ot_to_idx : dict
        Dictionary mapping object types to their id.
    bf_act : float
        Traversal will only explore this many enabled activities per step. If set, the play-out will generate a subset
        of all sequences. Will be stochastically rounded if not an integer.
    bf_bind : float
        Traversal will only explore this many enabled bindings per activity. If set, the play-out will generate a subset
        of all sequences. Will be stochastically rounded if not an integer.
    memo : dict
        The memoization cache where the state_key is mapped to a set of next steps or FINAL_MARKER if the state is the empty state.

    Returns
    -------
    bool
        Returns True if the state_key is reachable (i.e., not a deadlock), False otherwise.
        If the state_key is a deadlock, it will be represented by an empty set in the memo.
    """
    if state_key in memo:
        # a deadlock is indicated by an empty set in the memo.
        # an entry that is not empty indicates that the empty state is reachable
        return bool(memo[state_key])

    # state_key has not been explored yet, so we explore it
    state, activity_counts = state_key
    if not state.activities:
        # empty state
        memo[state_key] = FINAL_MARKER
        return True

    next_steps = set()
    enabled_activities = _get_enabled_activities(
        occn, semantics, state, start_activities, act_to_idx, idx_to_act, ot_to_idx
    )
    
    # Limit the number of enabled activities to bf_act
    if bf_act < float("inf"):
        # Stochastically round bf_act to an integer
        bf_act_rounded = int(bf_act) + (1 if random.random() < (bf_act % 1) else 0)
        # Select random subset of enabled activities
        enabled_activities = set(random.sample(list(enabled_activities), min(bf_act_rounded, len(enabled_activities))))

    # explore all sucessor states by binding all enabled activities
    for act in enabled_activities:
        act_id = act_to_idx[act]

        if activity_counts[act_id] >= max_bindings:
            continue

        new_activiy_counts = list(activity_counts)
        new_activiy_counts[act_id] += 1
        new_activity_counts_tuple = tuple(new_activiy_counts)

        # Get all enabled bindings for this activity
        if act_id in start_activities:
            enabled_bindings = _get_bindings_start_activity(
                occn, act, state, act_to_idx, ot_to_idx
            )
        else:
            enabled_bindings = semantics.enabled_bindings(occn, act, state, act_to_idx, ot_to_idx)

        # Limit the number of enabled bindings to bf_bind
        if bf_bind < float("inf"):
            # Stochastically round bf_bind to an integer
            bf_bind_rounded = int(bf_bind) + (1 if random.random() < (bf_bind % 1) else 0)
            # Select random subset of enabled bindings
            enabled_bindings = set(random.sample(list(enabled_bindings), min(bf_bind_rounded, len(enabled_bindings))))

        # explore all bindings
        for binding in enabled_bindings:
            new_state = semantics.bind_activity(
                occn,
                act=binding[0],
                cons=_convert_binding_tuple_to_dict(binding[1]),
                prod=_convert_binding_tuple_to_dict(binding[2]),
                state=state,
            )
            # clean up fake obligations for start activities
            if act_id in start_activities:
                new_state = _clean_fake_obligations(
                    occn, new_state, act, binding[2], act_to_idx, ot_to_idx
                )
            new_state_key = (new_state, new_activity_counts_tuple)

            if _populate_memo_graph(
                new_state_key,
                occn,
                semantics,
                max_bindings,
                start_activities,
                act_to_idx,
                idx_to_act,
                ot_to_idx,
                bf_act,
                bf_bind,
                memo
            ):
                next_steps.add((binding, new_state_key))

    # Add all next steps to memo
    # If state_key is a deadlock, this will be an empty set
    memo[state_key] = next_steps
    return bool(next_steps)


def _convert_binding_tuple_to_dict(binding_tuple):
    """
    Converts a tuple from a binding (conumed or produced) into a nested dictionary.
    None is converted to None.

    The inner values (object lists) are converted to sets.
    """
    if not binding_tuple:
        return None
    return {
        related_act: {
            object_type: set(objects) for object_type, objects in objects_per_type
        }
        for related_act, objects_per_type in binding_tuple
    }


def _get_enabled_activities(
    occn: OCCausalNet,
    semantics,
    state: OCCausalNetState,
    start_activities,
    act_to_idx: dict,
    idx_to_act: dict,
    ot_to_idx: dict,
) -> set:
    """
    Returns the enabled activities in the given state, including start activities
    if they have "fake obligations".

    Parameters
    -----------
    occn : OCCausalNet
        The causal net being used.
    semantics
        The semantics to be used for the causal net.
    state : OCCausalNetState
        The current state of the causal net.
    start_activities
        Collection of indices for start activities
    act_to_idx
        Dictionary mapping activities to their id.
    idx_to_act
        Dictionary mapping activity ids to their names.
    ot_to_idx
        Dictionary mapping object types to their id.

    Returns
    --------
    set
        A set of ids for enabled activities in the given state.
    """
    enabled_activities = set()

    # get start activities with outstanding fake obligations
    start_activities_with_obligations = state.activities.intersection(start_activities)
    # add names, not ids
    enabled_activities.update(idx_to_act[act_id] for act_id in start_activities_with_obligations)

    # get all other enabled activities
    enabled_activities.update(
        semantics.enabled_activities(
            occn,
            state,
            include_start_activities=False,
            act_to_idx=act_to_idx,
            ot_to_idx=ot_to_idx,
        )
    )

    return enabled_activities


def _get_bindings_start_activity(
    occn: OCCausalNet,
    act: str,
    state: OCCausalNetState,
    act_to_idx: dict,
    ot_to_idx: dict,
):
    """
    Computes all enabled bindings for a start activity with the given fake obligations
    in the state.

    Parameters
    -----------
    occn : OCCausalNet
        The object-centric causal net
    act : str
        The start activity to bind
    state : OCCausalNetState
        The current state of the causal net, which contains the fake obligations for the start activity.
    act_to_idx : dict
        Dictionary mapping activities to their id.
    ot_to_idx : dict
        Dictionary mapping object types to their id.

    Returns
    -----------
    tuple
        A tuple of enabled bindings for the start activity.
        Each binding is a tuple of (activity_id, consumed, produced).
        The consumed and produced are tuples of (predecessor/successor activity id, objects_per_ot),
        where objects_per_ot is a tuple of entries (object_type_id, objects).
    """
    # get the outstanding fake obligations for the start activity
    act_id = act_to_idx[act]
    obligations = state[act_id]
    if not obligations:
        return ()
    outstanding_objects = set()
    for (_, obj_id, _), _ in obligations.items():
        outstanding_objects.add(obj_id)

    # Extract object type
    object_type = act.split("_", 1)[1]

    # Compute enabled bindings
    bindings = OCCausalNetSemantics.enabled_bindings_start_activity(
        occn, act, object_type, outstanding_objects, act_to_idx, ot_to_idx
    )

    return bindings

def _clean_fake_obligations(
    occn: OCCausalNet,
    state: OCCausalNetState,
    act: str,
    produced: tuple,
    act_to_idx: dict,
    ot_to_idx: dict,
) -> OCCausalNetState:
    """
    Cleans up fake obligations for start activities in the state after binding the start activity.
    Since a firing of a start activity consumes no obligations,
    we need to manually remove the fake obligations that were created for the start activity
    for all objects that were bound to it.

    Parameters
    -----------
    occn : OCCausalNet
        The object-centric causal net.
    state : OCCausalNetState
        The current state of the causal net.
    act : str
        The activity that was bound.
    produced : tuple
        The produced tuple from the binding.
    act_to_idx : dict
        Dictionary mapping activities to their id.
    ot_to_idx : dict
        Dictionary mapping object types to their id.

    Returns
    -----------
    OCCausalNetState
        The updated state with cleaned fake obligations.
    """
    # get the set of all objects involved
    objects = set()
    object_types = set()
    for _, ot_to_obj in produced:
        for ot, obj_ids in ot_to_obj:
            objects.update(obj_ids)
            object_types.add(ot)
    
    assert len(object_types) == 1, "Only one object type should be involved in a start activity binding"
    ot_id = next(iter(object_types))

    act_id = act_to_idx[act]
    # remove all obligations for the start activity that are related to the objects
    state -= OCCausalNetState(
        {act_id: Counter([(-1, obj_id, ot_id) for obj_id in objects])}
    )
    
    return state


def _reconstruct_sequences(state_key: tuple, memo: dict):
    """
    Reconstructs valid binding sequences from the memoization cache.

    This function iterates over the memoization cache and reconstructs all valid binding sequences
    that lead to the empty state. It yields each sequence tuple of Binding objects.

    Parameters
    ----------
    state_key : tuple
        The key representing the current state in the memoization cache.
    memo : dict
        The memoization cache containing state keys and their corresponding next steps.

    Returns
    -------
    Iterator[tuple[Binding]]
        An iterator yielding tuples of bindings representing valid sequences.
    """
    next_steps = memo.get(state_key)

    if next_steps == FINAL_MARKER:
        # If we reached the empty state, yield an empty sequence
        yield ()
        return

    if not next_steps:
        # Deadlock state; this should only happen when there are 0 valid sequences
        # Do not yield anything
        return

    for binding, next_state_key in next_steps:
        # Recursively reconstruct sequences from the next state
        for sub_sequence in _reconstruct_sequences(next_state_key, memo):
            # Yield the current binding followed by the sub-sequence
            yield (binding,) + sub_sequence


def _valid_sequences_to_ocel(valid_sequences_iter, idx_to_act, idx_to_ot, parameters):
    """
    Converts the valid sequences of bindings into an OCEL object.

    Parameters
    ----------
    valid_sequences_iter : iter
        An iterator over valid sequences of bindings, where each sequence is a tuple of Binding objects
    idx_to_act : dict
        Mapping from indices to activity names
    idx_to_ot : dict
        Mapping from indices to object types
    parameters : dict
        Additional parameters for the conversion, including:
            Parameters.EVENT_ID: The column name for event IDs
            Parameters.OBJECT_ID: The column name for object IDs
            Parameters.OBJECT_TYPE: The column name for object types
            Parameters.EVENT_TIMESTAMP: The column name for event timestamps
            Parameters.EVENT_ACTIVITY: The column name for event activities

    Returns
    -------
    OCEL
        The resulting OCEL object.
    """
    event_id_column = exec_utils.get_param_value(
        Parameters.EVENT_ID, parameters, constants.DEFAULT_EVENT_ID
    )
    object_id_column = exec_utils.get_param_value(
        Parameters.OBJECT_ID, parameters, constants.DEFAULT_OBJECT_ID
    )
    object_type_column = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, constants.DEFAULT_OBJECT_TYPE
    )

    event_activity = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, constants.DEFAULT_EVENT_ACTIVITY
    )
    event_timestamp = exec_utils.get_param_value(
        Parameters.EVENT_TIMESTAMP, parameters, constants.DEFAULT_EVENT_TIMESTAMP
    )
    objects_unique_per_sequence = exec_utils.get_param_value(
        Parameters.OBJECTS_UNIQUE_PER_SEQUENCE, parameters, False
    )
    # Convert all found traces to OCEL format

    # Create the OCEL object
    events_list = []
    objects_list = []
    relations_list = []

    all_objects_seen = set()
    event_id_counter = 0
    # assigns to each event an increased timestamp from 1970
    curr_timestamp = 10000000
    
    if objects_unique_per_sequence:
        object_id_counter = 0

    for sequence in valid_sequences_iter:
        # For each sequence, create events and objects
        for binding in sequence:
            activity_id = binding[0]
            consumed = binding[1]
            produced = binding[2]

            act = idx_to_act[activity_id]

            # do not add START / END activities
            if act.startswith("START_") or act.startswith("END_"):
                continue

            # Create event
            event_id = f"event_{event_id_counter}"
            event_id_counter += 1
            curr_timestamp += 1

            events_list.append(
                {
                    event_id_column: event_id,
                    event_activity: act,
                    event_timestamp: pd.to_datetime(curr_timestamp, unit="s"),
                }
            )

            # Create objects and relations
            # consumed and produced contain the same objects; we only need to create them once
            for _, ot_to_obj in consumed:
                for ot_id, objects in ot_to_obj:
                    obj_type = idx_to_ot[ot_id]
                    for obj_id in objects:
                        if objects_unique_per_sequence:
                            obj_id = f"{obj_id}_{object_id_counter}"
                        
                        # Add object
                        if obj_id not in all_objects_seen:
                            all_objects_seen.add(obj_id)
                            objects_list.append(
                                {object_id_column: obj_id, object_type_column: obj_type}
                            )

                        # Add relation
                        relations_list.append(
                            {
                                event_id_column: event_id,
                                event_activity: act,
                                event_timestamp: pd.to_datetime(
                                    curr_timestamp, unit="s"
                                ),
                                object_id_column: obj_id,
                                object_type_column: obj_type,
                            }
                        )
        if objects_unique_per_sequence:
            object_id_counter += 1

    # Convert to dataframes
    events_df = pd.DataFrame(events_list)
    objects_df = pd.DataFrame(objects_list)
    relations_df = pd.DataFrame(relations_list)

    # Create the OCEL object
    ocel = OCEL(
        events=events_df,
        objects=objects_df,
        relations=relations_df,
        parameters=parameters,
    )

    return ocel



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from enum import Enum
from typing import Optional, Dict, Any, Union, List

import polars as pl

from pm4py.util import exec_utils, constants, xes_constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    CASE_ID_KEY = constants.PARAMETER_CONSTANT_CASEID_KEY
    SUBCASE_CONCAT_STR = "subcase_concat_str"
    CUT_MODE = "cut_mode"


def apply(
    df: pl.LazyFrame,
    activity: Union[str, List[str]],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> pl.LazyFrame:
    """
    Splits the cases of a log (Polars LazyFrame) into subcases based on the provision of an activity.
    There are as many subcases as many occurrences of a given activity occur.

    Example:
    Original log:

    [['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request', 'examine thoroughly',
    'check ticket', 'decide', 'pay compensation'],
    ['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request', 'check ticket',
    'examine casually', 'decide', 'reinitiate request', 'examine casually', 'check ticket', 'decide', 'reject request']]


    Log filtered using the activity split filter on 'reinitiate request' with cut_mode='this':

    [['register request', 'examine casually', 'check ticket', 'decide'],
    ['reinitiate request', 'examine thoroughly', 'check ticket', 'decide', 'pay compensation'],
    ['register request', 'examine casually', 'check ticket', 'decide'],
    ['reinitiate request', 'check ticket', 'examine casually', 'decide'],
    ['reinitiate request', 'examine casually', 'check ticket', 'decide', 'reject request']]


    Log filtered using the activity split filter on 'reinitiate request' with cut_mode='next':

    [['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request'],
    ['examine thoroughly', 'check ticket', 'decide', 'pay compensation'],
    ['register request', 'examine casually', 'check ticket', 'decide', 'reinitiate request'],
    ['check ticket', 'examine casually', 'decide', 'reinitiate request'],
    ['examine casually', 'check ticket', 'decide', 'reject request']]


    Parameters
    ----------------
    df
        LazyFrame
    activity
        Activity (or collection of activities)
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => activity key
        - Parameters.CASE_ID_KEY => case id
        - Parameters.SUBCASE_CONCAT_STR => concatenator between the case id and the subtrace index in the filtered df
        - Parameters.CUT_MODE => mode of cut:
            - "this" means that an event with the specified activity goes to the next subcase
            - "next" means that the following event (to the given activity) goes to the next subcase.

    Returns
    ----------------
    filtered_df
        LazyFrame in which the cases are split into subcases

    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )
    subcase_concat_str = exec_utils.get_param_value(
        Parameters.SUBCASE_CONCAT_STR, parameters, "##@@"
    )
    cut_mode = exec_utils.get_param_value(
        Parameters.CUT_MODE, parameters, "this"
    )

    activities = [activity] if isinstance(activity, str) else activity

    # Add row number within each case
    df = df.with_columns(
        pl.col(case_id_key).cum_count().over(case_id_key).alias("_row_within_case")
    )

    # Mark rows where the activity occurs
    df = df.with_columns(
        pl.col(activity_key).is_in(activities).alias("_is_split_activity")
    )

    # Calculate the subcase number for each event
    if cut_mode == "this":
        # Events with split activity start a new subcase
        df = df.with_columns(
            pl.col("_is_split_activity").cast(pl.Int32).cum_sum().over(case_id_key).alias("_subcase_idx")
        )
    else:  # cut_mode == "next"
        # Events after split activity start a new subcase
        df = df.with_columns(
            pl.col("_is_split_activity").shift(1, fill_value=False).over(case_id_key).alias("_shifted_split")
        )
        df = df.with_columns(
            pl.col("_shifted_split").cast(pl.Int32).cum_sum().over(case_id_key).alias("_subcase_idx")
        )
        df = df.drop("_shifted_split")

    # Create new case IDs by concatenating original case ID with subcase index
    df = df.with_columns(
        pl.concat_str([
            pl.col(case_id_key).cast(pl.Utf8),
            pl.lit(subcase_concat_str),
            pl.col("_subcase_idx").cast(pl.Utf8)
        ]).alias(case_id_key)
    )

    # Drop helper columns
    df = df.drop(["_row_within_case", "_is_split_activity", "_subcase_idx"])

    return df


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2026 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

# Author: Maximilian Josef Frank (https://orcid.org/0000-0002-0714-7748)

from collections import defaultdict
import copy
import itertools

from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.petri_net.utils import petri_utils


def apply(genetic_matrix, parameters=None):
    """
    Converts a Genetic matrix to a Petri net

    Reference paper:
    Maximilian Josef Frank. "Optimising and Implementing the Genetic Miner in PM4Py" (2026).

    Parameters
    --------------
    genetic_matrix
        Genetic matrix
    parameters
        (Unused) Possible parameters of the algorithm

    Returns
    --------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    # import here to resolve circular import dependencies
    from pm4py.algo.discovery.genetic.util import get_src_sink_sets_for_wfnet, iset
    I = genetic_matrix.input_map
    O = genetic_matrix.output_map
    T = genetic_matrix.transitions
    # @src 5.3. Sophisticated Mapping, Def. 7; https://doi.org/10.1007/11494744_5
    # original, but insanely inefficient (for |A|=23, |P(A)|=8388608):
    #    X = {(Ti, To) ∈ P(A) × P(A) | ∀t∈Ti To ∈ O(t) ∧ ∀t∈To Ti ∈ I(t)}
    # equivalent and more efficient (for |A|=23, |I(∀t)×O(∀t)|=24×16=384):
    #    X = {(Ti, To) ∈ I(∀t) × O(∀t) | ∀t∈Ti To ∈ O(t) ∧ ∀t∈To Ti ∈ I(t)}
    X = list(filter( # if no list(), X gets consumed as stream
        lambda Tio: # = (Ti, To)
            all(Tio[1] in O[t] for t in Tio[0]) and
            all(Tio[0] in I[t] for t in Tio[1]),
            # holds s. Def 4, constraint 3+4; https://doi.org/10.1007/11494744_5
        itertools.product(
            iset.flat(I.values()),
            iset.flat(O.values())
        )
    ))
    Xs = list(itertools.chain( # Xs = non-sophisticated / silent-transition places
        [ (Si,iset([t])) for Ti,t in zip(I.values(), I.keys()) for Si in Ti],
        [ (iset([t]),So) for t,To in zip(O.keys(), O.values()) for So in To]
    ))
    # remove places in Xs which are fully covered by X
    Xs = [ Tio for Tio in Xs if not any(
        Tio[0] <= Ti and Tio[1] <= To for Ti,To in X
    )]
    # remove Xs (silent transitions from non-simple matrix) from X (s. later)
    def is_transitively_not_sophisticated(Tio, places):
        """If place `Tio` is not in sophisticated mapping (i.e. place in `Xs`), all adjacent places (i.e. with same input/output transition) are replaced by non-sophisticated places and their silent transitions."""
        return any(
            len(Ti & Tio[0]) or len(To & Tio[1]) # & = intersect
            for Ti,To in places
        )
    check = copy.copy(X)
    while check:
        Tio = check.pop(0)
        if is_transitively_not_sophisticated(Tio, Xs):
            Xs.append(Tio)
            X.remove(Tio)
            for Tio2 in X:
                if Tio2 not in check and is_transitively_not_sophisticated(Tio2, Xs):
                    check.append(Tio2)
    # build PetriNet
    io = [PetriNet.Place('i'), PetriNet.Place('o')]
    net = PetriNet()
    net.places.update(PetriNet.Place(p) for p in map(str, X))
    net.places.update(io)
    # Non-silent transitions must have a label; label=None is rendered as invisible.
    net.transitions.update(PetriNet.Transition(t, t) for t in T)
    # original: F = {(i,t) | t∈T ∧ •t=∅} ∪ {(t,o) | t∈T ∧ t•=∅} ∪ …
    T_noI, T_noO = get_src_sink_sets_for_wfnet(I, O, T)
    arcs_pt = [ ('i', t) for t in T_noI ] + [
        ((Ti,To), t) for (Ti,To),t in itertools.product(X, T) if t in To
    ]
    arcs_tp = [ (t, 'o') for t in T_noO ] + [
        (t, (Ti,To)) for t,(Ti,To) in itertools.product(T, X) if t in Ti
    ]
    # convert to instances
    getPlace = lambda name: next(filter(lambda p: p.name == name, net.places), None)
    getTransition = lambda name: next(filter(lambda t: t.name == name, net.transitions), None)
    arcs_pt = [(getPlace(str(p)), getTransition(t)) for p,t in arcs_pt]
    arcs_tp = [(getTransition(t), getPlace(str(p))) for t,p in arcs_tp]
    # add instance arcs
    for i,o in itertools.chain(arcs_pt, arcs_tp):
        petri_utils.add_arc_from_to(i, o, net)
    # if causal matrix not simple (Def. 8; https://doi.org/10.1007/11494744_5):
    # @src 5.2. Naive Mapping, Def. 6; https://doi.org/10.1007/11494744_5
    # add incoming/outgoing silent places
    Ti = set(itertools.chain(*(Ti for Ti,_ in Xs)))
    To = set(itertools.chain(*(To for _,To in Xs)))
    Ps_o = { name : PetriNet.Place("o"+str(name))    # output of prev non-silent T
        for name in [ (t,s) for t in Ti for s in O[t] ]
    }
    Ps_i = { name : PetriNet.Place("i"+str(name))    # input of next non-silent T
        for name in [ (t,s) for t in To for s in I[t] ]
    }
    net.places.update(itertools.chain(Ps_o.values(), Ps_i.values()))
    Ts = defaultdict(lambda: PetriNet.Transition(name="")) # name="" → silent
    # add arcs
    for (t1,So),p in Ps_o.items():
        petri_utils.add_arc_from_to(getTransition(t1), p, net)
        for t2 in So: # connect silent T
            petri_utils.add_arc_from_to(p, Ts[t1,t2], net)
    for (t2,Si),p in Ps_i.items():
        petri_utils.add_arc_from_to(p, getTransition(t2), net)
        for t1 in Si: # connect silent T
            petri_utils.add_arc_from_to(Ts[t1,t2], p, net)
    net.transitions.update(Ts.values())
    # add markings
    markings = [Marking() for _ in io]
    for m,p in zip(markings, io):
        m[p] = 1
    return (net, *markings)



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from pm4py.objects.oc_causal_net.obj import OCCausalNet
import networkx as nx


def create_oc_causal_net(marker_groups):
    """
    Create an object-centric causal net from a list of marker groups.
    Does not consider activity counts or the relative occurence threshold.
    May mutate the input data.

    Parameters
    ----------
    marker_groups : dict[str, ]
        Dict of marker groups per activity. Syntax:
        {
            "activity_name": {
                "img": [
                    [
                        (activity, object_type, (min_count, max_count), marker_key),
                        // -1 for max_count = inf; 0 for unique marker key
                        ...
                    ],
                    ...
                ],
                "omg": [
                    ...
                ]
            }
            ]
        }

    Returns
    -------
    OCCausalNet
        Object-centric causal net
    """
    # infer activities
    activities = set(marker_groups.keys())

    # get input and output marker groups
    input_marker_groups = {}
    output_marker_groups = {}

    # make all keys=0 unique
    # find max key
    max_key = max(
        [
            key
            for groups in marker_groups.values()
            for group in groups.get("img", []) + groups.get("omg", [])
            for _, _, _, key in group
        ],
        default=0,
    )
    key_counter = max_key + 1

    # give markers with key=0 a unique key and set inf as max count if max count is -1
    for groups in marker_groups.values():
        for group in groups.get("img", []) + groups.get("omg", []):
            for i, (
                related_activity,
                object_type,
                count_range,
                marker_key,
            ) in enumerate(group):
                if marker_key == 0:
                    group[i] = (
                        related_activity,
                        object_type,
                        (
                            count_range
                            if count_range[1] != -1
                            else (count_range[0], float("inf"))
                        ),
                        key_counter,
                    )
                    key_counter += 1
            key_counter = max_key + 1

    for activity, groups in marker_groups.items():
        img = groups.get("img", [])
        omg = groups.get("omg", [])

        if img:
            input_marker_groups[activity] = [
                OCCausalNet.MarkerGroup(
                    markers=[
                        OCCausalNet.Marker(
                            related_activity, object_type, count_range, marker_key
                        )
                        for related_activity, object_type, count_range, marker_key in group
                    ]
                )
                for group in img
            ]
        if omg:
            output_marker_groups[activity] = [
                OCCausalNet.MarkerGroup(
                    markers=[
                        OCCausalNet.Marker(
                            related_activity, object_type, count_range, marker_key
                        )
                        for related_activity, object_type, count_range, marker_key in group
                    ]
                )
                for group in omg
            ]

    # infer arcs from the marker groups
    arcs = dict()
    for activity in activities:
        for group in output_marker_groups.get(activity, []):
            for marker in group.markers:
                related_activity = marker.related_activity
                object_type = marker.object_type
                if activity not in arcs:
                    arcs[activity] = {}
                if related_activity not in arcs[activity]:
                    arcs[activity][related_activity] = {}
                if object_type not in arcs[activity][related_activity]:
                    arcs[activity][related_activity][object_type] = {}
                arcs[activity][related_activity][object_type] = {
                    "object_type": object_type
                }
    # create the dependency graph
    dependency_graph = nx.MultiDiGraph(arcs)

    # create the object-centric causal net
    occn = OCCausalNet(
        dependency_graph,
        output_marker_groups,
        input_marker_groups,
    )
    return occn


'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
# Author: Maximilian Josef Frank (https://orcid.org/0000-0002-0714-7748)

from collections import defaultdict
import copy
import csv
from datetime import datetime
from pm4py.util.timeout import func_timeout, FunctionTimedOut
import numpy
import random
import itertools

try:
    from tqdm import tqdm
except ImportError:
    def tqdm(iterable=None, *args, **kwargs):
        return iterable if iterable is not None else []

import pm4py
from pm4py.util import exec_utils, constants
from pm4py.util import xes_constants as xes
from pm4py.algo.discovery.genetic.algorithm import Parameters
from pm4py.objects.conversion.log import converter as log_converter
from pm4py.objects.conversion.genetic_matrix.variants.to_petri_net import apply as matrix2petrinet
from pm4py.algo.discovery.genetic.util import get_src_sink_sets_for_wfnet, iset, rand_partition
from pm4py.objects.genetic_matrix.obj import GeneticMatrix

# typing
from typing import Any, Dict, Iterable, List, Optional, TextIO, Tuple, Union
from pandas.core.frame import DataFrame
from pm4py.objects.log.obj import EventLog, EventStream
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.algo.discovery.genetic.util import InputMap, OutputMap, Individual
from pm4py.utils import is_polars_lazyframe


def apply(
    log: Union[EventLog, EventStream, DataFrame],
    parameters: Optional[Dict[Union[str, Parameters], Any]] = None,
) -> Tuple[PetriNet, Marking, Marking]:
    """
    Discovers a Petri net using Genetic Miner

    Reference paper:
    Maximilian Josef Frank. "Optimising and Implementing the Genetic Miner in PM4Py" (2026).

    Parameters
    ------------
    log
        Event log
    parameters
        Possible parameters of the algorithm,
        including:
            - Parameters.ACTIVITY_KEY
            - Parameters.TIMESTAMP_KEY
            - Parameters.CASE_ID_KEY
            - Parameters.POPULATION_SIZE
            - Parameters.ELITISM_RATE
            - Parameters.CROSSOVER_RATE
            - Parameters.MUTATION_RATE
            - Parameters.GENERATIONS
            - Parameters.ELITISM_MIN_SAMPLE
            - Parameters.TOURNAMENT_TIMEOUT
            - Parameters.LOG_CSV

    Returns
    ------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    """
    if parameters is None:
        parameters = {}
    if is_polars_lazyframe(log):
        log = log.collect().to_pandas()
    if not isinstance(log, DataFrame):
        log = log_converter.apply(
            log,
            variant=log_converter.Variants.TO_DATA_FRAME,
            parameters=parameters,
        )

    activity_key = exec_utils.get_param_value(
        Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY
    )
    timestamp_key = exec_utils.get_param_value(
        Parameters.TIMESTAMP_KEY, parameters, xes.DEFAULT_TIMESTAMP_KEY
    )
    case_id_key = exec_utils.get_param_value(
        Parameters.CASE_ID_KEY, parameters, constants.CASE_CONCEPT_NAME
    )

    population_size = exec_utils.get_param_value(
        Parameters.POPULATION_SIZE, parameters, 500
    )
    elitism_rate = exec_utils.get_param_value(
        Parameters.ELITISM_RATE, parameters, 0.01
    )
    crossover_rate = exec_utils.get_param_value(
        Parameters.CROSSOVER_RATE, parameters, 1.0
    )
    mutation_rate = exec_utils.get_param_value(
        Parameters.MUTATION_RATE, parameters, 0.01
    )
    generations = exec_utils.get_param_value(
        Parameters.GENERATIONS, parameters, 100
    )
    elitism_min_sample = exec_utils.get_param_value(
        Parameters.ELITISM_MIN_SAMPLE, parameters, 5
    )
    tournament_timeout = exec_utils.get_param_value(
        Parameters.TOURNAMENT_TIMEOUT, parameters, None
    )
    log_csv = exec_utils.get_param_value(
        Parameters.LOG_CSV, parameters, None
    )

    # configure parameters
    if population_size < 2:
        raise ValueError("population_size < 2: You need at least two parents for each next generation, thus at least a population size of 2.")
    if elitism_min_sample >= population_size:
        elitism_min_sample = population_size - 1
    if elitism_min_sample < 1:
        raise ValueError("elitism_min_sample < 1: No empty samples allowed.")
    if log_csv:
        log_csv = csv.writer(log_csv)
        log_csv.writerow(['timestamp', 'generation'] + [f'fitness{i}' for i in range(population_size)])
    # @src 6.3. Genetic Operations; https://doi.org/10.1007/11494744_5
    T = tuple(log[activity_key].unique())
    if tournament_timeout is None or tournament_timeout < 1:
        tournament_timeout = 3 * len(T) * log.shape[0] * (46 / 1202267 / 26)
        # based on previous runs: 46s for 1202267 traces with 26 activities
    if tournament_timeout < 1:
        tournament_timeout = 1
    history = []
    best_individual = None
    best_fitness = -1.0
    population = individuals(log, population_size, T, { # [(I,O), …]
        "activity_key":activity_key, "timestamp_key":timestamp_key, "case_id_key":case_id_key
    })
    for _ in tqdm(range(generations), "Genetic generations"):
        population, fitness = tournament(
            tqdm(population, f"└─Tournament {len(history)}"),
            log,
            T,
            sort=True,
            timeout=tournament_timeout,
            activity_key=activity_key,
            timestamp_key=timestamp_key,
            case_id_key=case_id_key,
        )
        if fitness[0] > best_fitness:
            best_fitness = fitness[0]
            best_individual = copy.deepcopy(population[0])
        if log_csv:
            log_csv.writerow([datetime.now(), len(history)] + fitness)
        if fitness[0] == 1 or (history and all(f == fitness[0] for f in history[-int(generations/2):])):
            break
        history.append(fitness[0])
        # elitism
        next_population = population[:round(elitism_rate * population_size)]
        # fill with offsprings
        while len(next_population) < population_size:
            parent1, parent2 = sample_parents(population, fitness, elitism_min_sample)
            if random.random() < crossover_rate:
                offsprings = crossover(parent1, parent2, T)
            else:
                offsprings = copy.deepcopy((parent1, parent2))
            for offspring in offsprings:
                if len(next_population) < population_size:
                    offspring = mutate(offspring, mutation_rate)
                    next_population.append(offspring)
        population = next_population
    if best_individual is None:
        population, fitness = tournament(
            population,
            log,
            T,
            sort=True,
            timeout=tournament_timeout,
            activity_key=activity_key,
            timestamp_key=timestamp_key,
            case_id_key=case_id_key,
        )
        best_individual = copy.deepcopy(population[0])
    return matrix2petrinet(GeneticMatrix(*best_individual, T))

def individuals(log: Union[DataFrame, EventLog], sample_size=1, T=None, keys: Dict[str, str] = {"activity_key":xes.DEFAULT_NAME_KEY, "timestamp_key":xes.DEFAULT_TIMESTAMP_KEY, "case_id_key":constants.CASE_CONCEPT_NAME}) -> List[Individual]:
    if not T:
        T = tuple(log[keys['activity_key']].unique())
    T_idx = {activity: idx for idx, activity in enumerate(T)}
    # @src 6.1. Initial Population; https://doi.org/10.1007/11494744_5
    # create matrix C
    C = numpy.zeros((len(T), len(T)))
    for _,group in tqdm(log.sort_values(keys['timestamp_key'], ascending=True).groupby(keys['case_id_key']), desc="Find consecutive activities"):
        prev_activity = None
        for _, row in group.iterrows():
            activity = row[keys['activity_key']]
            if prev_activity is not None:
                C[T_idx[prev_activity], T_idx[activity]] += 1
            prev_activity = activity
    row_sums = C.sum(axis=1, keepdims=True)
    Cn = numpy.divide(C, row_sums, out=numpy.zeros_like(C, dtype=float), where=row_sums != 0)
    samples = []
    for _ in tqdm(range(sample_size), "Initial population"):
        I,O = defaultdict(list), defaultdict(list)
        for i,o in numpy.ndindex(C.shape):
            if random.random() < Cn[i,o]:    # [0,1[ < [0,1]; 0 < 0 = false
                O[T[i]].append(T[o])
                I[T[o]].append(T[i])
        I,O = repair(I, O, C, T)
        # partitioning already ensures no T in >1 partitions
        # s. 4. Causal Matrix, Def. 4; https://doi.org/10.1007/11494744_5
        for i in I.keys():
            I[i] = rand_partition(I[i])
        for o in O.keys():
            O[o] = rand_partition(O[o])
        samples.append((I,O))
    return samples

def tournament(
    population: List[Individual],
    log: Union[DataFrame, EventLog],
    T,
    sort=True,
    timeout=1,
    *,
    activity_key: str = xes.DEFAULT_NAME_KEY,
    timestamp_key: str = xes.DEFAULT_TIMESTAMP_KEY,
    case_id_key: str = constants.CASE_CONCEPT_NAME,
) -> Tuple[List[Individual], List[float]]:
    """sort=True: sort descending by fitness (i.e. best first)"""
    # @src 6.2. Fitness Calculation; https://doi.org/10.1007/11494744_5
    fitness = []
    for i,(I,O) in enumerate(population):
        model = matrix2petrinet(GeneticMatrix(I,O,T))
        try:
            metrics = func_timeout(
                timeout = timeout,
                func = pm4py.fitness_token_based_replay,
                args = (log, *model),
                kwargs = {
                    "activity_key": activity_key,
                    "timestamp_key": timestamp_key,
                    "case_id_key": case_id_key,
                },
            )
        except FunctionTimedOut:
            print("\tTimeout for individual", i)
            metrics = defaultdict(lambda: 0)
        fitness.append(
            0.4 * metrics['average_trace_fitness']
            # @see https://pm4py-source.readthedocs.io/en/stable/_modules/pm4py/algo/evaluation/replay_fitness/variants/token_replay.html#evaluate
            # average_fitness = sum_fitness / #traces
            # = ∑ fitness per trace / #traces
            #   ∑ fitness per trace * len(log)/#traces
            # = ––––––––––––––––––––––––––––––––––––––
            #        #traces * len(log)/#traces
            #   ∑ fitness per row in log   ∑ fitness per activity
            # = –––––––––––––––––––––––– = ––––––––––––––––––––––
            #           len(log)                 len(log)
            +
            0.6 * metrics['percentage_of_fitting_traces']/100
        )
    if sort:
        population, fitness = zip(*sorted(
            zip(population, fitness),
            key=lambda obj: obj[1],
            reverse=True
        ))
        population, fitness = list(population), list(fitness)
    return (population, fitness)

def sample_parents(population: List[Individual], fitness: List[float], elitism_min_sample: int):
    population_fitness = tuple(zip(population, fitness))
    parent1 = sorted(
        random.sample(population_fitness, k=elitism_min_sample),
        key=lambda obj: obj[1],
        reverse=True
    )[0][0]
    parent2 = sorted(
        random.sample(
            [(i,f) for i,f in population_fitness if i!=parent1],
            k=elitism_min_sample
        ),
        key=lambda obj: obj[1],
        reverse=True
    )[0][0]
    return (parent1, parent2)

def crossover(parent1: Individual, parent2: Individual, T: List[str]) -> Tuple[Individual, Individual]:
    # @src 6.3. Genetic Operations: Crossover; https://doi.org/10.1007/11494744_5
    # 1. cross-over point
    t = random.choice(T)
    # 2. clone parents
    I1,O1 = offspring1 = copy.deepcopy(parent1)
    I2,O2 = offspring2 = copy.deepcopy(parent2)
    # 3. swap and recombine
    if I1[t] and I2[t]:
        old_I1 = iset.flat(I1[t])
        old_I2 = iset.flat(I2[t])
        swap_point = random.randrange(min(len(I1[t]), len(I2[t])))
        toI1, toI2 = I2[t][swap_point:], I1[t][swap_point:]
        # no T can exist twice in I/O[t], s. Def. 4; https://doi.org/10.1007/11494744_5
        # COPY of I_i, else not properly removed in opposite I_j
        I1_flat = iset.flat(I1[t][:swap_point])
        toI1_dedup = [iset(S - I1_flat) for S in toI1 if S - I1_flat]
        I2_flat = iset.flat(I2[t][:swap_point])
        toI2_dedup = [iset(S - I2_flat) for S in toI2 if S - I2_flat]
        # merge
        I1[t], I2[t] = I1[t][:swap_point] + toI1_dedup, I2[t][:swap_point] + toI2_dedup
        new_I1 = iset.flat(I1[t])
        new_I2 = iset.flat(I2[t])
        for c in new_I1 - old_I1:
            _add_singleton_partition(O1, c, t)
        for c in old_I1 - new_I1:
            _remove_value_from_partitions(O1, c, t)
        for c in new_I2 - old_I2:
            _add_singleton_partition(O2, c, t)
        for c in old_I2 - new_I2:
            _remove_value_from_partitions(O2, c, t)
    if O1[t] and O2[t]:
        old_O1 = iset.flat(O1[t])
        old_O2 = iset.flat(O2[t])
        swap_point = random.randrange(min(len(O1[t]), len(O2[t])))
        toO1, toO2 = O2[t][swap_point:], O1[t][swap_point:]
        # no T can exist twice in I/O[t], s. Def. 4; https://doi.org/10.1007/11494744_5
        # COPY of I_i, else not properly removed in opposite I_j
        O1_flat = iset.flat(O1[t][:swap_point])
        toO1_dedup = [iset(S - O1_flat) for S in toO1 if S - O1_flat]
        O2_flat = iset.flat(O2[t][:swap_point])
        toO2_dedup = [iset(S - O2_flat) for S in toO2 if S - O2_flat]
        # merge
        O1[t], O2[t] = O1[t][:swap_point] + toO1_dedup, O2[t][:swap_point] + toO2_dedup
        new_O1 = iset.flat(O1[t])
        new_O2 = iset.flat(O2[t])
        for c in new_O1 - old_O1:
            _add_singleton_partition(I1, c, t)
        for c in old_O1 - new_O1:
            _remove_value_from_partitions(I1, c, t)
        for c in new_O2 - old_O2:
            _add_singleton_partition(I2, c, t)
        for c in old_O2 - new_O2:
            _remove_value_from_partitions(I2, c, t)
    return (offspring1, offspring2)


def _add_singleton_partition(mapping: Dict[str, List[iset]], key: str, value: str) -> None:
    partitions = mapping.setdefault(key, [])
    if any(value in partition for partition in partitions):
        return
    partitions.append(iset({value}))


def _remove_value_from_partitions(mapping: Dict[str, List[iset]], key: str, value: str) -> None:
    partitions = mapping.setdefault(key, [])
    for index, partition in enumerate(partitions):
        if value in partition:
            updated = iset(partition - {value})
            if updated:
                partitions[index] = updated
            else:
                del partitions[index]
            return

def mutate(individual: Individual, rate: float = 0.01) -> Individual:
    # @src 6.3. Genetic Operations: Mutation; https://doi.org/10.1007/11494744_5
    I,O = individual
    for t in I.keys():
        if random.random() < rate:
            I[t] = rand_partition(itertools.chain(*I[t]))
    for t in O.keys():
        if random.random() < rate:
            O[t] = rand_partition(itertools.chain(*O[t]))
    return (I,O)

def repair(I: InputMap, O: OutputMap, C: numpy.ndarray, T: List[str]) -> Individual:
    """
    Merging partitions until petri-net is (coherent) workflow net
    s. last requirement in 4. Causal Matrix, Def. 4; https://doi.org/10.1007/11494744_5
    """
    # WF-Net Definition: each node on path from i to o
    #   → re-connect partitions of each I and O, not I and O together
    left = set(T)
    partitions = []
    while left:
        partition = set()
        Tn = {left.pop()}
        while Tn:
            t = Tn.pop() # also removes from Tn
            left.discard(t) # remove without Exception (s. initial Tn)
            partition.add(t)
            Tn |= (set(I[t]) | set(O[t])) & left
        partitions.append(partition)
    T_idx = {activity: idx for idx, activity in enumerate(T)}

    def rand_connect_one(I: InputMap, O: OutputMap, Ti: Iterable[str], To: Iterable[str], C: numpy.ndarray) -> Individual:
        comb = tuple(itertools.product(Ti, To))
        try:
            ti,to = random.choices(
                comb,
                weights = [ C[T_idx[ti], T_idx[to]] for ti,to in comb ],
                k=1
            )[0]
        except ValueError:
            ti,to = random.choice(comb)
        O[ti].append(to)
        I[to].append(ti)
        return I,O
    while len(partitions) > 1:
        random.shuffle(partitions)
        # merge Ti / To of partition to other partition
        first, last = get_src_sink_sets_for_wfnet(I, O, partitions[1])
        for i in first:
            # connect any→i
            I,O = rand_connect_one(I, O, [i], partitions[0], C)
        for o in last:
            # connect o→any
            I,O = rand_connect_one(I, O, partitions[0], [o], C)
        # pop first + merge with new first (i.e. old second)
        merging = partitions.pop(0)
        partitions[0] |= merging
    return (I,O)



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.conversion.heuristics_net.variants import to_petri_net
from enum import Enum
from pm4py.util import exec_utils


class Variants(Enum):
    TO_PETRI_NET = to_petri_net


def apply(heu_net, parameters=None, variant=Variants.TO_PETRI_NET):
    """
    Converts an Heuristics Net to a different type of object

    Parameters
    --------------
    heu_net
        Heuristics net
    parameters
        Possible parameters of the algorithm
    variant
        Variant of the algorithm:
            - Variants.TO_PETRI_NET
    """
    return exec_utils.get_variant(variant).apply(
        heu_net, parameters=parameters
    )



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
import sys
import uuid
from enum import Enum
from pm4py.util import exec_utils, constants
import tempfile
from graphviz import Digraph
from pm4py.util import vis_utils
from typing import Dict, Optional, Any, Tuple


class Parameters(Enum):
    FORMAT = "format"
    BGCOLOR = "bgcolor"
    ACTIVITY_THRESHOLD = "activity_threshold"
    EDGE_THRESHOLD = "edge_threshold"
    AGGREGATION_MEASURE = "aggregation_measure"
    ENABLE_GRAPH_TITLE = "enable_graph_title"
    GRAPH_TITLE = "graph_title"


def apply(
    network_analysis_edges0: Dict[Tuple[str, str], Dict[str, Any]],
    parameters: Optional[Dict[Any, Any]] = None,
) -> Digraph:
    """
    Creates a visualization of the network analysis (performance view)

    Parameters
    -----------------
    network_analysis_edges0
        Edges of the network analysis
    parameters
        Parameters of the algorithm, including:
        - Parameters.FORMAT => the format of the visualization
        - Parameters.BGCOLOR => the background color
        - Parameters.ACTIVITY_THRESHOLD => the minimum number of occurrences for an activity to be included (default: 1)
        - Parameters.EDGE_THRESHOLD => the minimum number of occurrences for an edge to be included (default: 1)
        - Parameters.AGGREGATION_MEASURE => the aggregation measure (default: mean)

    Returns
    ------------------
    digraph
        Graphviz graph
    """
    if parameters is None:
        parameters = {}

    image_format = exec_utils.get_param_value(
        Parameters.FORMAT, parameters, "png"
    )
    bgcolor = exec_utils.get_param_value(
        Parameters.BGCOLOR, parameters, constants.DEFAULT_BGCOLOR
    )
    activity_threshold = exec_utils.get_param_value(
        Parameters.ACTIVITY_THRESHOLD, parameters, 1
    )
    edge_threshold = exec_utils.get_param_value(
        Parameters.EDGE_THRESHOLD, parameters, 1
    )
    aggregation_measure = exec_utils.get_param_value(
        Parameters.AGGREGATION_MEASURE, parameters, "mean"
    )

    from statistics import mean, median, stdev

    aggregation_f = mean

    if aggregation_measure == "median":
        aggregation_f = median
    elif aggregation_measure == "min":
        aggregation_f = min
    elif aggregation_measure == "max":
        aggregation_f = max
    elif aggregation_measure == "stdev":
        aggregation_f = stdev
    elif aggregation_measure == "sum":
        aggregation_f = sum

    enable_graph_title = exec_utils.get_param_value(
        Parameters.ENABLE_GRAPH_TITLE,
        parameters,
        constants.DEFAULT_ENABLE_GRAPH_TITLES,
    )
    graph_title = exec_utils.get_param_value(
        Parameters.GRAPH_TITLE, parameters, "Network Analysis"
    )

    filename = tempfile.NamedTemporaryFile(suffix=".gv")
    filename.close()

    viz = Digraph(
        "pt",
        filename=filename.name,
        engine="dot",
        graph_attr={"bgcolor": bgcolor},
    )
    viz.attr("node", shape="ellipse", fixedsize="false")

    if enable_graph_title:
        viz.attr(
            label='<<FONT POINT-SIZE="20">' + graph_title + "</FONT>>",
            labelloc="top",
        )

    network_analysis_edges = {}
    network_analysis_edges_agg_performance = {}
    for x in network_analysis_edges0:
        network_analysis_edges[x] = {}
        network_analysis_edges_agg_performance[x] = {}
        for y in network_analysis_edges0[x]:
            network_analysis_edges[x][y] = len(network_analysis_edges0[x][y])
            network_analysis_edges_agg_performance[x][y] = aggregation_f(
                network_analysis_edges0[x][y]
            )

    nodes = set(x[0] for x in network_analysis_edges).union(
        set(x[1] for x in network_analysis_edges)
    )
    nodes_in_degree = {x: 0 for x in nodes}
    nodes_out_degree = {x: 0 for x in nodes}
    for edge in network_analysis_edges:
        for edge_value in network_analysis_edges[edge]:
            if network_analysis_edges[edge][edge_value] >= edge_threshold:
                nodes_in_degree[edge[1]] += network_analysis_edges[edge][
                    edge_value
                ]
                nodes_out_degree[edge[0]] += network_analysis_edges[edge][
                    edge_value
                ]
    nodes_max_degree = {
        x: max(nodes_in_degree[x], nodes_out_degree[x]) for x in nodes
    }

    max_node_value = sys.maxsize
    min_node_value = -sys.maxsize

    nodes_dict = {}
    for node in nodes_max_degree:
        if nodes_max_degree[node] >= activity_threshold:
            nodes_dict[node] = str(uuid.uuid4())
            viz.node(
                nodes_dict[node],
                node
                + "\n(in="
                + str(nodes_in_degree[node])
                + "; out="
                + str(nodes_out_degree[node])
                + ")",
                style="filled",
                fillcolor=vis_utils.get_trans_freq_color(
                    nodes_max_degree[node], max_node_value, max_node_value
                ),
            )
            count = nodes_max_degree[node]
            if count > max_node_value:
                max_node_value = count
            elif count < min_node_value:
                min_node_value = count

    min_edge_value = sys.maxsize
    max_edge_value = -sys.maxsize

    for edge in network_analysis_edges:
        if edge[0] in nodes_dict and edge[1] in nodes_dict:
            for edge_value in network_analysis_edges[edge]:
                count = network_analysis_edges[edge][edge_value]
                if count > max_edge_value:
                    max_edge_value = count
                elif count < min_edge_value:
                    min_edge_value = count

    for edge in network_analysis_edges:
        if edge[0] in nodes_dict and edge[1] in nodes_dict:
            for edge_value in network_analysis_edges[edge]:
                if network_analysis_edges[edge][edge_value] >= edge_threshold:
                    viz.edge(
                        nodes_dict[edge[0]],
                        nodes_dict[edge[1]],
                        label=edge_value
                        + "\n"
                        + vis_utils.human_readable_stat(
                            network_analysis_edges_agg_performance[edge][
                                edge_value
                            ]
                        )
                        + "",
                        penwidth=str(
                            vis_utils.get_arc_penwidth(
                                network_analysis_edges[edge][edge_value],
                                min_edge_value,
                                max_edge_value,
                            )
                        ),
                    )

    viz.format = image_format.replace("html", "plain-ext")

    return viz



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Set



def pre_set(occn, activity: str, object_type: str = None) -> Set:
    """
    Returns the set of predecessor activities for a given activity in an object-centric causal net.
    Restricted to predecessors connected using arcs of the specified object type, if provided.
    
    Parameters
    ----------
    occn : OCCausalNet
        The object-centric causal net to query
    activity : str
        The name of the activity for which to get the predecessors
    object_type : str, optional
        The object type to restrict the predecessors to (default is None)
    
    Returns
    -------
    Set
        Set of predecessor activities of the specified object type.
    """
    if activity not in occn.activities:
        return set()
    
    dg = occn.dependency_graph
    
    if object_type is None:
        return dg.predecessors(activity)
    
    return {
        predecessor
        for predecessor, _, edge_key in dg.in_edges(activity, keys=True)
        if edge_key == object_type
    }

def post_set(occn, activity: str, object_type: str = None) -> Set:
    """
    Returns the set of successor activities for a given activity in an object-centric causal net.
    Restricted to successors connected using arcs of the specified object type, if provided.
    
    Parameters
    ----------
    occn : OCCausalNet
        The object-centric causal net to query
    activity : str
        The name of the activity for which to get the successors
    object_type : str, optional
        The object type to restrict the successors to (default is None)
    
    Returns
    -------
    Set
        Set of successor activities of the specified object type.
    """
    if activity not in occn.activities:
        return set()
    
    dg = occn.dependency_graph
    
    if object_type is None:
        return dg.successors(activity)
    
    return {
        successor
        for _, successor, edge_key in dg.out_edges(activity, keys=True)
        if edge_key == object_type
    }
    
    
    
    


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from abc import ABC, abstractmethod
from typing import Generic, TypeVar, Tuple, List, Optional, Dict, Any

from pm4py.algo.discovery.inductive.dtypes.im_ds import IMDataStructure
from pm4py.objects.process_tree.obj import ProcessTree

T = TypeVar("T", bound=IMDataStructure)


class FallThrough(ABC, Generic[T]):

    @classmethod
    @abstractmethod
    def holds(cls, t: T, parameters: Optional[Dict[str, Any]] = None) -> bool:
        pass

    @classmethod
    @abstractmethod
    def apply(
        cls,
        t: T,
        pool=None,
        manager=None,
        parameters: Optional[Dict[str, Any]] = None,
    ) -> Optional[Tuple[ProcessTree, List[T]]]:
        pass



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from threading import Thread
from typing import Any, Callable, Dict, Optional, Tuple
import ctypes


class FunctionTimedOut(TimeoutError):
    def __init__(
        self,
        timeout: float,
        func: Callable[..., Any],
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.timeout = timeout
        self.func = func
        self.args = args if args is not None else ()
        self.kwargs = kwargs if kwargs is not None else {}
        func_name = getattr(func, "__name__", repr(func))
        super().__init__(f"Function '{func_name}' timed out after {timeout} seconds")


def _request_thread_stop(thread: Thread) -> None:
    ident = thread.ident
    if ident is None:
        return

    try:
        set_async_exc = ctypes.pythonapi.PyThreadState_SetAsyncExc
    except Exception:
        return

    set_async_exc.argtypes = [ctypes.c_ulong, ctypes.py_object]
    set_async_exc.restype = ctypes.c_int
    result = set_async_exc(ctypes.c_ulong(ident), ctypes.py_object(SystemExit))
    if result > 1:
        set_async_exc(ctypes.c_ulong(ident), None)


def func_timeout(
    timeout: float,
    func: Callable[..., Any],
    args: Optional[Tuple[Any, ...]] = None,
    kwargs: Optional[Dict[str, Any]] = None,
) -> Any:
    if timeout is None:
        return func(*(args or ()), **(kwargs or {}))
    if timeout <= 0:
        raise ValueError("timeout must be > 0")

    args = args if args is not None else ()
    kwargs = kwargs if kwargs is not None else {}

    result: Dict[str, Any] = {}
    error: Dict[str, BaseException] = {}

    def _runner() -> None:
        try:
            result["value"] = func(*args, **kwargs)
        except BaseException as exc:
            error["value"] = exc

    thread = Thread(target=_runner, daemon=True)
    thread.start()
    thread.join(timeout)

    if thread.is_alive():
        _request_thread_stop(thread)
        thread.join(0.05)
        raise FunctionTimedOut(timeout, func, args=args, kwargs=kwargs)

    if "value" in error:
        raise error["value"]

    return result.get("value")



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

import os
import sqlite3
from datetime import datetime
from enum import Enum
from typing import Optional, Dict, Any

import pandas as pd
from pm4py.util.dt_parsing.variants import strpfromiso
from pm4py.util import exec_utils, pandas_utils


class Parameters(Enum):
    """Enum container for optional *apply* parameters."""

    HISTORY_DB_PATH = "history_db_path"


# -----------------------------------------------------------------------------
# Firefox-internal transition type constants → human-readable labels
# -----------------------------------------------------------------------------
VISIT_TYPE_MAP = {
    0: "LINK",  # Followed a hyperlink (default navigation)
    1: "TYPED",  # URL typed in the address bar (or pasted & Go)
    2: "BOOKMARK",  # Visit launched from a bookmark
    3: "EMBED",  # Automatically loaded in an embedded context (img/iframe)
    4: "REDIRECT_PERM",  # 301 permanent redirect target
    5: "REDIRECT_TEMP",  # 302/307 temporary redirect target
    6: "DOWNLOAD",  # Explicit file download
    7: "FRAMED_LINK",  # Link inside a frameset / iframe clicked by user
    8: "RELOAD",  # Page reload (F5 / Ctrl-R)
    9: "USER_TYPED",  # Legacy typed constant (alias of 1)
}


# -----------------------------------------------------------------------------
# Public API
# -----------------------------------------------------------------------------

def apply(parameters: Optional[Dict[Any, str]] = None) -> pd.DataFrame:
    """Extract an event log from one or more *places.sqlite* files.

    The function can be pointed **either** at a specific *places.sqlite* file
    **or** at the *Profiles* directory containing multiple profile folders. If
    a directory is given, a separate case identifier is generated for every
    profile found.

    Attributes created per event (column → description)
    --------------------------------------------------
    * **case:concept:name** – Identifier of the Firefox profile folder (or
      "DEFAULT" when a single DB path is supplied). Serves as *case ID* in
      process-mining terminology.

    * **concept:name** – Canonicalised URL path (protocol/query stripped).
      Acts as *activity label* for each event.

    * **time:timestamp** – Python ``datetime`` converted from Firefox's PRTime
      (microseconds since Unix epoch).

    * **complete_url** – Full URL including scheme and query string, suitable
      when the exact navigation target is important.

    * **domain** – Registered domain portion of the URL; helps grouping visits
      by site.

    * **url_wo_parameters** – Same as *concept:name* but kept separately for
      convenience when the activity label gets renamed downstream.

    * **title** – Page title stored by Firefox (from ``moz_places.title``).
      Note: may be *NULL* for some automatic loads or download links.

    * **visit_type** – Raw integer from ``moz_historyvisits.visit_type``.
      Referential value for advanced analysis.

    * **visit_type_desc** – Human-friendly label mapped from ``visit_type`` via
      :data:`VISIT_TYPE_MAP` (e.g., "TYPED", "BOOKMARK", "RELOAD").

    * **session_id** – Numeric identifier grouping visits that happened within
      the same browser session (i.e., between Firefox startups).

    * **from_visit** – ``visit_id`` of the *referrer* visit inside the same
      database – enables reconstruction of click chains within a tab.

    * **visit_count_place** – Cumulative counter how often the *place* (URL)
      has been visited overall up to the current record.

    * **typed** – ``True`` when the user explicitly typed/pasted the address
      (``moz_places.typed > 0``). Often correlates with intentional navigations.

    * **frecency** – Firefox’s scoring metric combining *frequency* and
      *recency*; higher numbers indicate pages likely to appear as top results
      in the address bar.

    * **guid** – Stable 12-character base-62 GUID uniquely identifying the
      *place* entry – useful as a surrogate key when merging logs.

    Returns
    -------
    pandas.DataFrame
        Chronologically sorted event log ready for process-mining libraries.
    """
    if parameters is None:
        parameters = {}

    # ------------------------------------------------------------------
    # Resolve DB location (explicit path or default Windows profile dir)
    # ------------------------------------------------------------------
    history_db_path = exec_utils.get_param_value(
        Parameters.HISTORY_DB_PATH,
        parameters,
        os.path.join(
            os.path.expanduser("~"),
            "AppData",
            "Roaming",
            "Mozilla",
            "Firefox",
            "Profiles",
        ),
    )

    # Collect one (file path, profile name) tuple per profile
    if os.path.isdir(history_db_path):
        profiles = [
            (os.path.join(history_db_path, p, "places.sqlite"), p)
            for p in os.listdir(history_db_path)
        ]
    else:
        profiles = [(history_db_path, "DEFAULT")]

    profiles = [p for p in profiles if os.path.exists(p[0])]

    # ------------------------------------------------------------------
    # Unified SQL pulling all the attributes listed above
    # ------------------------------------------------------------------
    QUERY = (
        "SELECT "
        "p.url, v.visit_date, p.title, v.visit_type, v.session, v.from_visit, "
        "p.visit_count, p.typed, p.frecency, p.guid "
        "FROM moz_historyvisits v JOIN moz_places p ON v.place_id = p.id"
    )

    events = []

    for db_path, profile_name in profiles:
        try:
            conn = sqlite3.connect(db_path)
            curs = conn.cursor()
            curs.execute(QUERY)

            rows = curs.fetchall()
            for (
                url,
                visit_date,
                title,
                visit_type,
                session_id,
                from_visit,
                visit_count_place,
                typed,
                frecency,
                guid,
            ) in rows:
                # Convert micro-second PRTime → datetime (naïve, local time)
                ts = strpfromiso.fix_naivety(datetime.fromtimestamp(visit_date / 1_000_000))

                ev = {
                    "case:concept:name": profile_name,
                    "concept:name": url.split("//")[-1].split("?")[0].replace(",", ""),
                    "complete_url": url,
                    "domain": url.split("//")[-1].split("/")[0],
                    "url_wo_parameters": url.split("//")[-1].split("?")[0],
                    "time:timestamp": ts,
                    "title": title,
                    "visit_type": visit_type,
                    "visit_type_desc": VISIT_TYPE_MAP.get(visit_type, "UNKNOWN"),
                    "session_id": session_id,
                    "from_visit": from_visit,
                    "visit_count_place": visit_count_place,
                    "typed": bool(typed),
                    "frecency": frecency,
                    "guid": guid,
                }

                if ev["case:concept:name"].strip() and ev["concept:name"].strip():
                    events.append(ev)
        finally:
            curs.close()
            conn.close()

    # ------------------------------------------------------------------
    # Convert list → DataFrame and impose stable, chronological ordering
    # ------------------------------------------------------------------
    df = pandas_utils.instantiate_dataframe(events)

    if not df.empty:
        df = pandas_utils.insert_index(df, "@@index", copy_dataframe=False, reset_index=False)
        df = df.sort_values(["time:timestamp", "@@index"], kind="mergesort")
        df["@@case_index"] = df.groupby("case:concept:name", sort=False).ngroup()
        df = df.sort_values(["@@case_index", "time:timestamp", "@@index"], kind="mergesort")

    return df



'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.objects.ocel import constants
from enum import Enum
from pm4py.util import exec_utils
import datetime as dt
import numpy as np
import pandas as pd


class Parameters(Enum):
    EVENT_ACTIVITY = constants.PARAM_EVENT_ACTIVITY
    OBJECT_TYPE = constants.PARAM_OBJECT_TYPE


def _infer_type_from_series(series):
    dtype_str = str(series.dtype)
    if "date" in dtype_str or "time" in dtype_str:
        return "date"
    if "float" in dtype_str or "double" in dtype_str:
        return "float"
    if "int" in dtype_str:
        return "int"
    if "bool" in dtype_str:
        return "boolean"

    if dtype_str != "object":
        return "string"

    values = series.dropna()
    if values.empty:
        return "string"

    def _isinstance_all(values_iter, types):
        return all(isinstance(v, types) for v in values_iter)

    values_list = values.tolist()
    if _isinstance_all(values_list, (bool,)):
        return "boolean"
    if _isinstance_all(values_list, (int,)):
        return "int"
    if _isinstance_all(values_list, (float, int)):
        return "float"
    if _isinstance_all(values_list, (pd.Timestamp, np.datetime64, dt.datetime, dt.date)):
        return "date"

    return "string"


def get(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    if parameters is None:
        parameters = {}

    event_activity_column = exec_utils.get_param_value(
        Parameters.EVENT_ACTIVITY, parameters, ocel.event_activity
    )
    object_type_column = exec_utils.get_param_value(
        Parameters.OBJECT_TYPE, parameters, ocel.object_type_column
    )

    ets = {}
    for k, v in ocel.events.groupby(event_activity_column):
        cols = v.dropna(axis="columns", how="all").columns
        ets[k] = {
            x: _infer_type_from_series(v[x])
            for x in cols
            if not x.startswith("ocel:")
        }

    ots = {}
    for k, v in ocel.objects.groupby(object_type_column):
        cols = v.dropna(axis="columns", how="all").columns
        ots[k] = {
            x: _infer_type_from_series(v[x])
            for x in cols
            if not x.startswith("ocel:")
        }

    ots2 = {}
    for k, v in ocel.object_changes.groupby(object_type_column):
        cols = v.dropna(axis="columns", how="all").columns
        ots2[k] = {
            x: _infer_type_from_series(v[x])
            for x in cols
            if not x.startswith("ocel:")
        }

    for k in ots2:
        if k not in ots:
            ots[k] = ots2[k]
        else:
            for x in ots2[k]:
                if x not in ots[k]:
                    ots[k][x] = ots2[k][x]

    return ets, ots



'''
    PM4Py â€“ A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschrÃ¤nkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''
# Author: Maximilian Josef Frank (https://orcid.org/0000-0002-0714-7748)

import random
import itertools

# typing
from collections.abc import Iterable
from typing import Dict, FrozenSet, List, Tuple
InputMap = Dict[str, List[FrozenSet]]
OutputMap = Dict[str, List[FrozenSet]]
Individual = Tuple[InputMap, OutputMap]

class iset(frozenset):
    "Indexable frozenset printing as set, i.e. without `frozenset(…)`"
    def __repr__(self):
        return "{" + repr(sorted(self))[1:-1] + "}"

    @staticmethod
    def flat(item: Iterable) -> "iset":
        return iset(itertools.chain(*item))

def rand_partition(pool: Iterable) -> List[iset]:
    pool = set(pool)
    #     also ensures no activity in two partitions
    #     s. 4. Causal Matrix, Def. 4; https://doi.org/10.1007/11494744_5
    partition = []
    while pool:
        draw = iset(random.sample(
            tuple(pool),
            random.randint(1, len(pool))
        ))
        partition.append(draw)
        pool -= draw
    return partition

def get_src_sink_sets_for_wfnet(I: InputMap, O: OutputMap, T: List[str]) -> Tuple[List[str], List[str]]:
    """Determines input set and output set, which need to be connected by a place to create a WF-net"""
    def add2graphs(graphs, t, nextT):
        # find graph
        graph = next((g for g in graphs if t in g), None)
        if graph is None:
            graph = [t]
            graphs.append(graph)
        # append plain/grouped T
        successors = []
        for S in nextT[t]:
            if type(S) == str:
                successors.append(S)
            else:
                successors.extend(S)
        graph += successors
        # merge if end = start
        for tn in successors:
            for g2 in graphs[:]: # [:] = copy
                if g2[0] == tn and g2 != graph:
                    graph += g2
                    graphs.remove(g2)
                    break
        return graphs
    graphsI, graphsO = [], []
    for t in T:
        graphsI = add2graphs(graphsI, t, I)
        graphsO = add2graphs(graphsO, t, O)
    first = [g[0] for g in graphsO] # ⋃first = reachable via O[∀t]
    last = [g[0] for g in graphsI] # ⋃first = reachable via I[∀t]
    return first, last



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from typing import Any, Dict
from pm4py.objects.ocpn.obj import OCMarking, OCPetriNet
from pm4py.objects.petri_net.obj import Marking, PetriNet
from pm4py.objects.petri_net.utils.petri_utils import add_arc_from_to


def apply(ocpn: OCPetriNet, parameters=None) -> Dict[str, Any]:
    """
    Converts an OCPetriNet object to the alternative format as a Dict[str, Any] specified in pm4py.algo.discovery.ocel.ocpn.variants.classic.
    Only the essential components of the OCPetriNet are retained in the alternative format: `activities`, `object_types`, `petri_nets`, `double_arcs_on_activity`, `start_activities`, and `end_activities`.

    Parameters
    ----------
    ocpn: OCPetriNet
        The object-centric Petri net to be converted.
    parameters: dict, optional
        Additional parameters for the conversion (not used in this implementation).

    Returns
    ----------
    ocpn: Dict[str, Any]
        Alternative format representation of the object-centric Petri net.
    """
    if parameters is None:
        parameters = {}

    object_types = ocpn.object_types
    activities = {t.label for t in ocpn.transitions if t.label}
    petri_nets = {ot: _project_ocpn_on_object_type(ocpn, ot) for ot in object_types}
    double_arcs_on_activity = _get_double_arcs(ocpn)
    start_activities = _get_start_end_activities(ocpn, ocpn.initial_marking)
    end_activities = _get_start_end_activities(ocpn, ocpn.final_marking)

    alternative_format = {
        "activities": activities,
        "object_types": object_types,
        "petri_nets": petri_nets,
        "double_arcs_on_activity": double_arcs_on_activity,
        "start_activities": start_activities,
        "end_activities": end_activities,
        # information not extracted in this implementation
        "edges": {
            "event_couples": {ot: {} for ot in object_types},
            "event_pairs": {ot: {} for ot in object_types},
            "total_objects": {ot: {} for ot in object_types},
        },
        "activities_indep": {
            "events": {ot: {} for ot in object_types},
            "unique_objects": {ot: {} for ot in object_types},
            "total_objects": {ot: {} for ot in object_types},
        },
        "activities_ot": {
            "events": {ot: {} for ot in object_types},
            "unique_objects": {ot: {} for ot in object_types},
            "total_objects": {ot: {} for ot in object_types},
        },
        "tbr_results": {},
    }

    return alternative_format


def _project_ocpn_on_object_type(
    ocpn: OCPetriNet, object_type
) -> tuple[PetriNet, Marking, Marking]:
    """
    Projects the OCPetriNet into a tuple containing the Petri net projection and the initial and final marking projections for the object type.

    Parameters
    ----------
    ocpn: OCPetriNet
        The object-centric Petri net to be split.
    object_type: str
        The object type for which the projection is to be created.

    Returns
    ----------
    tuple[PetriNet, Marking, Marking]
        A tuple containing the Petri net projection, initial marking, and final marking projection for the object type.
    """
    # extract places by ot
    places = [p for p in ocpn.places if p.object_type == object_type]

    # extract arcs from those places
    arcs = [
        a
        for p in places
        for a in p.out_arcs | p.in_arcs
        if a.object_type == object_type
    ]

    # extract transitions as those used in the arcs
    transitions = {
        a.source for a in arcs if isinstance(a.source, OCPetriNet.Transition)
    } | {a.target for a in arcs if isinstance(a.target, OCPetriNet.Transition)}

    # construct net projection
    pn_places = {p: PetriNet.Place(p.name) for p in places}
    pn_transitions = {t: PetriNet.Transition(name=t.name, label=t.label) for t in transitions}

    # create pn
    pn = PetriNet(
        name=f"{ocpn.name}_{object_type}",
        places=set(pn_places.values()),
        transitions=set(pn_transitions.values()),
    )

    # add arcs
    for arc in arcs:
        source = pn_places.get(arc.source, pn_transitions.get(arc.source))
        target = pn_places.get(arc.target, pn_transitions.get(arc.target))

        add_arc_from_to(source, target, pn)

    # initial (& final) marking as multiset of places in the initial marking where the count is the number of objects of that type in the place
    initial_marking = oc_marking_to_petri(ocpn.initial_marking, pn_places)
    final_marking = oc_marking_to_petri(ocpn.final_marking, pn_places)

    return pn, initial_marking, final_marking


def oc_marking_to_petri(
    oc_marking: OCMarking,
    pn_places: Dict[OCPetriNet.Place, PetriNet.Place],
) -> Marking:
    """
    Convert an object-centric marking to a classic Petri-net Marking that
    contains only the places that are keys of the `pn_places` dictionary.

    Parameters
    ----------
    oc_marking: OCMarking
        The object-centric marking to convert.
    pn_places: Dict[OCPetriNet.Place, PetriNet.Place]
        A mapping from object-centric Petri net places to classic Petri net places.

    Returns
    ----------
    Marking
        A classic Petri net marking (place -> token count)
    """
    petri_marking = Marking()
    if not oc_marking:
        return petri_marking

    # Aggregate multiplicities per place for the requested object type
    for place, counter in oc_marking.items():
        if place in pn_places.keys():
            if pn_places[place] not in petri_marking:
                petri_marking[pn_places[place]] = 0
            petri_marking[pn_places[place]] += sum(counter.values())

    return petri_marking


def _get_double_arcs(ocpn: OCPetriNet) -> Dict[str, Any]:
    """
    Returns a dictionary mapping each object type to a dict mapping an activity to True if only connected to variable arcs, or False if only connected to non-variable arcs.

    Parameters
    ----------
    ocpn: OCPetriNet
        The object-centric Petri net to analyze.

    Returns
    ----------
    Dict[str, Any]
        A dictionary where keys are object types and values are dicts where keys are
        activity names and values are True if only connected to variable arcs, or False if only connected to non-variable arcs.
    """
    double_arcs = {ot: {} for ot in ocpn.object_types}
    for arc in ocpn.arcs:
        ot = arc.object_type
        act = arc.source.label if isinstance(arc.source, OCPetriNet.Transition) else arc.target.label
        if act is None:
            continue
        if act in double_arcs[ot]:
            if double_arcs[ot][act] != arc.is_variable:
                raise ValueError(
                    f"Transition {act} in object type {ot} is connected to both variable and non-variable arcs. The given OCPetriNet is invalid."
                )
        double_arcs[ot][act] = arc.is_variable

    return double_arcs


def _get_start_end_activities(
    ocpn: OCPetriNet,
    marking: OCMarking
) -> Dict[str, Any]:
    """
    Returns a dictionary mapping each object type to a dict mapping the start/end activities to empty sets for events, unique_objects, and total_objects.

    Parameters
    ----------
    ocpn: OCPetriNet
        The object-centric Petri net to analyze.
    marking: OCMarking
        The initial or final marking of the OCPetriNet, used to determine start or end activities.

    Returns
    ----------
    Dict[str, Any]
        The start or end activities dictionaries.
    """
    # activities are those occuring in the given marking
    activities = {ot: {} for ot in ocpn.object_types}
    
    if marking is None:
        return activities
    
    for p in marking.places:
        ot = p.object_type
        if p not in activities[ot]:
            activities[ot][p.name] = {
                "events": set(),
                "unique_objects": set(),
                "total_objects": set(),
            }
            
    return activities


'''
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
'''

from typing import Optional, Dict, Any
from pm4py.objects.petri_net.obj import PetriNet, Marking
from enum import Enum
from pm4py.util import exec_utils


class Parameters(Enum):
    RESPONSE_HEADER = "response_header"


def apply(
    net: PetriNet,
    im: Marking,
    fm: Marking,
    parameters: Optional[Dict[Any, Any]] = None,
) -> str:
    """
    Provides the description of an accepting Petri net

    Parameters
    --------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    parameters
        Possible parameters of the algorithm, including:
        - Parameters.INCLUDE_HEADER => includes the header

    Returns
    --------------
    stru
        String representation of the given accepting Petri net
    """
    if parameters is None:
        parameters = {}

    include_header = exec_utils.get_param_value(
        Parameters.RESPONSE_HEADER, parameters, True
    )

    ret = ["\n"]
    if include_header:
        ret.append("If I have a Petri net:\n")
    ret.append(repr(net))
    ret.append("\ninitial marking: " + repr(im))
    ret.append("final marking: " + repr(fm))
    ret.append("\n")

    return "\n".join(ret)



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""

from collections import Counter, defaultdict
from copy import deepcopy
from typing import Collection, Dict, Any, Set
from pm4py.objects.petri_net.obj import PetriNet


class OCMarking(defaultdict):
    """
    An object-centric marking represented as a mapping from places to multisets of object IDs.
    
    ```
    marking = OCMarking({p: Counter(["object1", "object2"])})
    ```
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initializes the OCMarking, querying unspecified places defaults to an empty multiset."""
        super().__init__(Counter)
        data_args = args
        if args and args[0] is Counter:
            data_args = args[1:]
        initial_data = dict(*data_args, **kwargs)
        for place, objects in initial_data.items():
            self[place] = Counter(objects)

    def __hash__(self):
        return frozenset(
            (place, frozenset(counter.items()))
            for place, counter in self.items()
            if counter
        ).__hash__()

    def __eq__(self, other):
        if not isinstance(other, OCMarking):
            return False
        return all(
            self.get(p, Counter()) == other.get(p, Counter())
            for p in set(self.keys()) | set(other.keys())
        )

    def __le__(self, other):
        for place, self_counter in self.items():
            other_counter = other.get(place, Counter())
            # Every object count in self must be less than or equal to the count in other.
            if not all(
                other_counter.get(obj_id, 0) >= count
                for obj_id, count in self_counter.items()
            ):
                return False
        return True

    def __add__(self, other):
        result = OCMarking()
        for place, self_counter in self.items():
            result[place] += self_counter
        for place, other_counter in other.items():
            result[place] += other_counter
        return result

    def __sub__(self, other):
        result = OCMarking()
        for place, self_counter in self.items():
            diff = self_counter - other.get(place, Counter())
            if diff != Counter():
                result[place] = diff
        return result

    def __repr__(self):
        # e.g.  ["p1:{o1, o2}", "p2: {o2, o3}", …]
        sorted_entries = sorted(self.items(), key=lambda item: item[0].name)
        return (
            ", ".join(f"{place.name}: {objects}" for (place, objects) in sorted_entries)
            if sorted_entries
            else "[]"
        )

    def __str__(self):
        return self.__repr__()

    def __deepcopy__(self, memodict={}):
        new_marking = OCMarking()
        memodict[id(self)] = new_marking
        for place, objects in self.items():
            place_copy = (
                memodict[id(place)]
                if id(place) in memodict
                else deepcopy(place, memodict)
            )
            counter_copy = (
                memodict[id(objects)]
                if id(objects) in memodict
                else deepcopy(objects, memodict)
            )
            new_marking[place_copy] = counter_copy
        return new_marking

    @property
    def places(self) -> Set:
        """
        Returns the set of all places in the marking that contain tokens.

        Returns
        ------------
        Set[str]
            Set of place names in the marking.
        """
        return set([p for p in self.keys() if self[p]])


class OCPetriNet(PetriNet):
    class Place(PetriNet.Place):
        def __init__(
            self, name, object_type, in_arcs=None, out_arcs=None, properties=None
        ):
            """
            Constructor

            Parameters
            ------------
            name
                human-readable identifier
            object_type
                the type/color of objects this place holds
            in_arcs
                set of incoming arcs
            out_arcs
                set of outgoing arcs
            properties
                dict of additional properties
            """
            super().__init__(
                name, in_arcs=in_arcs, out_arcs=out_arcs, properties=properties
            )
            self.__object_type = object_type

        def add_in_arc(self, arc):
            """
            Adds an incoming arc to the place.

            Parameters
            ------------
            arc: OCPetriNet.Arc
                the arc to add
            """
            self.__in_arcs.add(arc)
            assert arc.target == self
            assert arc.object_type == self.object_type

        def add_out_arc(self, arc):
            """
            Adds an outgoing arc to the place.

            Parameters
            ------------
            arc: OCPetriNet.Arc
                the arc to add
            """
            self.__out_arcs.add(arc)
            assert arc.source == self
            assert arc.object_type == self.object_type

        def __get_object_type(self):
            return self.__object_type

        def __repr__(self):
            return f"{self.name}[{self.object_type}]"

        def __deepcopy__(self, memodict={}):
            if id(self) in memodict:
                return memodict[id(self)]
            new_place = OCPetriNet.Place(
                self.name, self.object_type, properties=self.properties
            )
            memodict[id(self)] = new_place
            # attached arcs
            for arc in self.in_arcs:
                arc_copy = deepcopy(arc, memodict)
                new_place.in_arcs.add(arc_copy)
            for arc in self.out_arcs:
                arc_copy = deepcopy(arc, memodict)
                new_place.out_arcs.add(arc_copy)

            return new_place

        object_type = property(__get_object_type)

    class Transition(PetriNet.Transition):
        def add_in_arc(self, arc):
            """
            Adds an incoming arc to the place.

            Parameters
            ------------
            arc: OCPetriNet.Arc
                the arc to add
            """
            self.__in_arcs.add(arc)
            assert arc.target == self

        def add_out_arc(self, arc):
            """
            Adds an outgoing arc to the place.

            Parameters
            ------------
            arc: OCPetriNet.Arc
                the arc to add
            """
            self.__out_arcs.add(arc)
            assert arc.source == self
            

    class Arc(PetriNet.Arc):
        def __init__(
            self,
            source,
            target,
            object_type,
            is_variable=False,
            properties=None,
        ):
            """
            Constructor

            Parameters
            ------------
            source
                source place / transition
            target
                target place / transition
            is_variable
                whether the arc is a variable arc
            properties
                dict of additional properties
            """
            super().__init__(source, target, weight=1, properties=properties)
            self.__object_type = object_type
            self.__is_variable = is_variable

        def __get_object_type(self):
            return self.__object_type

        def __get_is_variable(self):
            return self.__is_variable

        def __repr__(self):
            base = super().__repr__()
            var = "variable" if self.is_variable else "non-variable"
            return f"{base}:{self.object_type}:{var}"

        def __deepcopy__(self, memodict={}):
            if id(self) in memodict:
                return memodict[id(self)]
            new_source = memodict.get(id(self.source), deepcopy(self.source, memodict))
            new_target = memodict.get(id(self.target), deepcopy(self.target, memodict))
            new_arc = OCPetriNet.Arc(
                new_source,
                new_target,
                self.object_type,
                is_variable=self.is_variable,
                properties=self.properties,
            )
            memodict[id(self)] = new_arc
            # reattach
            new_source.out_arcs.add(new_arc)
            new_target.in_arcs.add(new_arc)
            return new_arc

        object_type = property(__get_object_type)
        is_variable = property(__get_is_variable)

    def __init__(
        self,
        name: str = None,
        places: Collection[Place] = None,
        transitions: Collection[Transition] = None,
        arcs: Collection[Arc] = None,
        initial_marking: OCMarking = None,
        final_marking: OCMarking = None,
        properties: Dict[str, Any] = None,
    ):
        """
        Constructor

        Parameters
        ------------
        name
            human-readable identifier
        places
            collection of places
        transitions
            collection of transitions
        arcs
            collection of arcs
        initial_marking
            initial marking of the net
        final_marking
            final marking of the net
        properties
            dict of additional properties
        """
        super().__init__(
            name=name,
            places=places,
            transitions=transitions,
            arcs=arcs,
            properties=properties,
        )
        self.__initial_marking = initial_marking
        self.__final_marking = final_marking
        self.__assert_well_formed()

    def __get_initial_marking(self):
        return self.__initial_marking

    def __get_final_marking(self):
        return self.__final_marking

    def __deepcopy__(self, memodict={}):
        new_net = OCPetriNet(self.name)
        memodict[id(self)] = new_net
        for p in self.places:
            p_copy = OCPetriNet.Place(p.name, p.object_type, properties=p.properties)
            new_net.places.add(p_copy)
            memodict[id(p)] = p_copy
        for t in self.transitions:
            t_copy = OCPetriNet.Transition(t.name, t.label, properties=t.properties)
            new_net.transitions.add(t_copy)
            memodict[id(t)] = t_copy
        for a in self.arcs:
            src = memodict[id(a.source)]
            tgt = memodict[id(a.target)]
            a_copy = OCPetriNet.Arc(
                src,
                tgt,
                a.object_type,
                is_variable=a.is_variable,
                properties=a.properties,
            )
            src.out_arcs.add(a_copy)
            tgt.in_arcs.add(a_copy)
            new_net.arcs.add(a_copy)
            memodict[id(a)] = a_copy

        def copy_marking(marking):
            if marking is None:
                return None
            copied_marking = OCMarking()
            for place, objects in marking.items():
                place_copy = memodict.get(id(place))
                if place_copy is None:
                    place_copy = OCPetriNet.Place(
                        place.name, place.object_type, properties=place.properties
                    )
                    memodict[id(place)] = place_copy
                    new_net.places.add(place_copy)
                copied_marking[place_copy] = objects.copy()
            return copied_marking

        new_net._OCPetriNet__initial_marking = copy_marking(self.initial_marking)
        new_net._OCPetriNet__final_marking = copy_marking(self.final_marking)
        return new_net

    def __repr__(self):
        ret = [f"OCPN {self.name}:\nobject_types: ["]
        object_types_rep = []
        for ot in self.object_types:
            object_types_rep.append(ot)
        object_types_rep.sort()
        ret.append(" " + ", ".join(object_types_rep) + " ")
        ret.append("]\nplaces: [")
        places_rep = []
        for place in self.places:
            places_rep.append(repr(place))
        places_rep.sort()
        ret.append(" " + ", ".join(places_rep) + " ")
        ret.append("]\ntransitions: [")
        trans_rep = []
        for trans in self.transitions:
            trans_rep.append(repr(trans))
        trans_rep.sort()
        ret.append(" " + ", ".join(trans_rep) + " ")
        ret.append("]\narcs: [")
        arcs_rep = []
        for arc in self.arcs:
            arcs_rep.append(repr(arc))
        arcs_rep.sort()
        ret.append(" " + ", ".join(arcs_rep) + " ")
        ret.append("]\ninitial_marking: [")
        initial_marking_rep = [repr(self.initial_marking)]
        ret.append(" " + ", ".join(initial_marking_rep) + " ")
        ret.append("]\nfinal_marking: [")
        final_marking_rep = [repr(self.final_marking)]
        ret.append(" " + ", ".join(final_marking_rep) + " ")
        ret.append("]")
        return "".join(ret)

    def __str__(self):
        return self.__repr__()

    def __assert_well_formed(self):
        """
        Asserts that the OCPN is well-formed, i.e., all transitions have,
        for each object type, only either variable or non-variable arcs, but not both.
        """
        for t in self.transitions:
            for ot in self.object_types:
                var_arcs = {
                    a for a in t.in_arcs if a.is_variable and a.object_type == ot
                }
                non_var_arcs = {
                    a for a in t.in_arcs if not a.is_variable and a.object_type == ot
                }
                if var_arcs and non_var_arcs:
                    raise ValueError(f"Transition {t} is not well-formed.")

    initial_marking = property(__get_initial_marking)
    final_marking = property(__get_final_marking)

    @property
    def object_types(self) -> Set[str]:
        """
        Returns the set of all object types (colors) used in this net.

        Returns
        ------------
        Set[str]
            Set of object types (colors) used in this net.
        """
        return {p.object_type for p in self.places}



"""
    PM4Py – A Process Mining Library for Python
Copyright (C) 2024 Process Intelligence Solutions UG (haftungsbeschränkt)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as
published by the Free Software Foundation, either version 3 of the
License, or any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see this software project's root or
visit <https://www.gnu.org/licenses/>.

Website: https://processintelligence.solutions
Contact: info@processintelligence.solutions
"""


def filter4(input_marker_groups, output_marker_groups, threshold, activity_count):
    """
    Filters input_marker_groups and output_marker_groups by relative support, then
    recursively keeps only those MarkerGroups that remain connected.

    Parameters
    ----------
    input_marker_groups : Dict[str, List[OCCausalNet.MarkerGroup]]
        Input marker groups of the activities
    output_marker_groups : Dict[str, List[OCCausalNet.MarkerGroup]]
        Output marker groups of the activities
    threshold : float
        Minimum relative support = support_count / activity_count[activity]
    activity_count : Dict[str,int]
        Absolute frequency of each activity in the event log.

    Returns
    -------
    Tuple[
        Dict[str, List[OCCausalNet.MarkerGroup]],
        Dict[str, List[OCCausalNet.MarkerGroup]]
    ]
        (filtered_input_marker_groups, filtered_output_marker_groups)
    """

    def filterByTreshold(marker_groups, activity):
        """
        Filters a list of MarkerGroups by relative support.

        Parameters
        ----------
        marker_groups : List[OCCausalNet.MarkerGroup]
            The MarkerGroups to filter.
        activity : str
            Activity name whose frequency is used for relative support.

        Returns
        -------
        List[OCCausalNet.MarkerGroup]
            Those groups whose support_count / activity_count[activity] > threshold.
        """
        filteredMarkerGroups = [
            marker_group
            for marker_group in marker_groups
            if marker_group.support_count / activity_count[activity] > threshold
        ]
        return filteredMarkerGroups

    def getMostFrequent(marker_groups):
        """
        Returns the most frequent MarkerGroup in the list.

        Parameters
        ----------
        marker_groups : List[OCCausalNet.MarkerGroup]
            List of marker groups to examine.

        Returns
        -------
        OCCausalNet.MarkerGroup
            The one with the highest support_count.
        """
        try:
            most_frequent_marker_group = marker_groups[
                [marker_group.support_count for marker_group in marker_groups].index(
                    max([marker_group.support_count for marker_group in marker_groups])
                )
            ]
        except ValueError as e:
            raise Exception(f"Invalid marker groups provided for some activity. Error: {e}")
        return most_frequent_marker_group

    def getSubsequentInputMarkerGroups(most_frequent_marker_group, activity):
        """
        For every marker in *most_frequent_marker_group* (which belongs to an
        **output** marker group of *activity*) look up the corresponding input
        marker groups of the marker's successor activity.
        Keeps only the most frequent ones and adds them to the
        *filtered_input_marker_groups* structure.

        Parameters
        ----------
        most_frequent_marker_group : OCCausalNet.MarkerGroup
        activity : str
        """
        for marker in most_frequent_marker_group.markers:
            succ_activity = marker.related_activity
            if succ_activity in input_marker_groups.keys():
                possible_input_marker_groups = [
                    x
                    for x in input_marker_groups[succ_activity]
                    if (activity, marker.object_type)
                    in [(y.related_activity, y.object_type) for y in x.markers]
                ]
                most_frequent_marker_group = getMostFrequent(possible_input_marker_groups)
                addToFilteredInputMarkerGroups(most_frequent_marker_group, succ_activity)

    def addToFilteredOutputMarkerGroups(most_frequent_marker_group, activity):
        """
        Inserts *most_frequent_marker_group* into *filtered_output_marker_groups* and
        triggers the recursive traversal to the input side.

        Parameters
        ----------
        most_frequent_marker_group : OCCausalNet.MarkerGroup
        activity : str
        """
        if most_frequent_marker_group not in filtered_output_marker_groups[activity]:
            filtered_output_marker_groups[activity].append(most_frequent_marker_group)
            getSubsequentInputMarkerGroups(most_frequent_marker_group, activity)

    def getSubsequentOutputMarkerGroups(most_frequent_marker_group, activity):
        """
        For every marker in *most_frequent_marker_group* (which belongs to an
        **input** marker_group of *activity*) look up the corresponding output
        marker groups of the marker's predecessor activity.
        Keeps only the most frequent ones and adds them to the
        *filtered_output_marker_groups* structure.

        Parameters
        ----------
        most_frequent_marker_group : OCCausalNet.MarkerGroup
        activity : str
        """
        for marker in most_frequent_marker_group.markers:
            pred_activity = marker.related_activity
            if pred_activity in output_marker_groups.keys():
                possible_output_marker_groups = [
                    x
                    for x in output_marker_groups[pred_activity]
                    if (activity, marker.object_type)
                    in [(y.related_activity, y.object_type) for y in x.markers]
                ]
                most_frequent_marker_group = getMostFrequent(possible_output_marker_groups)
                addToFilteredOutputMarkerGroups(most_frequent_marker_group, pred_activity)

    def addToFilteredInputMarkerGroups(most_frequent_marker_group, activity):
        """
        Inserts *most_frequent_marker_group* into *filtered_input_marker_groups* and
        triggers the recursive traversal to the output side.

        Parameters
        ----------
        most_frequent_marker_group : OCCausalNet.MarkerGroup
        activity : str
        """
        if most_frequent_marker_group not in filtered_input_marker_groups[activity]:
            filtered_input_marker_groups[activity].append(most_frequent_marker_group)
            getSubsequentOutputMarkerGroups(most_frequent_marker_group, activity)

    filtered_output_marker_groups = {act: [] for act in output_marker_groups.keys()}
    filtered_input_marker_groups = {act: [] for act in input_marker_groups.keys()}

    for act in filtered_output_marker_groups.keys():
        filteredMarkerGroups = filterByTreshold(output_marker_groups[act], act)
        for marker_group in filteredMarkerGroups:
            addToFilteredOutputMarkerGroups(marker_group, act)

    for act in filtered_input_marker_groups.keys():
        filteredMarkerGroups = filterByTreshold(input_marker_groups[act], act)
        for marker_group in filteredMarkerGroups:
            addToFilteredInputMarkerGroups(marker_group, act)

    return filtered_input_marker_groups, filtered_output_marker_groups
